# Copyright (c) 2018-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Custom pytest fixtures for GWpy.

This module is imported in gwpy.conftest such that all fixtures declared
here are available to test functions/methods by default.

Developer note: **none of the fixtures here should declare autouse=True**.
"""

from __future__ import annotations

import os
from typing import TYPE_CHECKING
from unittest import mock

import numpy
import pytest
from matplotlib import rc_context

from ..plot.tex import has_tex
from ..timeseries import TimeSeries
from . import mocks

if TYPE_CHECKING:
    from collections.abc import Iterator

    import nds2

# -- plotting ------------------------

SKIP_TEX: pytest.MarkDecorator = pytest.mark.skipif(
    not has_tex(),
    reason="TeX is not available",
)


@pytest.fixture(params=[
    pytest.param(False, id="no-tex"),
    pytest.param(True, id="usetex", marks=SKIP_TEX),
])
def usetex(
    request: pytest.FixtureRequest,
) -> Iterator[bool]:
    """Repeat a test with ``text.usetex`` `False` and `True`.

    If TeX is not available on the test machine (determined by
    :func:`gwpy.plot.tex.has_tex`), the ``usetex=True`` tests will be skipped.
    """
    use_ = request.param
    with rc_context(rc={"text.usetex": use_}):
        yield use_


# -- various useful mock data series -

_FIXTURE_TS_SAMPLE_RATE = 2048.
_FIXTURE_TS_DURATION = 10.  # seconds
_FIXTURE_TS_NSAMPLES = int(_FIXTURE_TS_SAMPLE_RATE * _FIXTURE_TS_DURATION)

@pytest.fixture
def white_noise() -> TimeSeries:
    """10s of white noise, sampled at 2kHz, starting at t0=0s."""
    rng = numpy.random.default_rng(seed=0)
    noise = rng.normal(size=_FIXTURE_TS_NSAMPLES)
    return TimeSeries(
        noise / numpy.std(noise),  # Scale to 1V/rtHz
        dt=1 / _FIXTURE_TS_SAMPLE_RATE,
        t0=0,
        name="noise",
    )

@pytest.fixture
def colored_noise(white_noise: TimeSeries) -> TimeSeries:
    """10s of 1/f^2 coloured noise, sampled at 2kHz, starting at t0=0s."""
    freqs = numpy.fft.rfftfreq(_FIXTURE_TS_NSAMPLES, 1/_FIXTURE_TS_SAMPLE_RATE)

    # Create coloured noise
    noise_fft = numpy.fft.rfft(white_noise)
    coloring_filter = numpy.ones_like(freqs)
    coloring_filter[1:] = 1 / (freqs[1:]**2)
    coloring_filter[0] = coloring_filter[1]  # Handle DC
    colored_noise = numpy.fft.irfft(noise_fft * coloring_filter, n=_FIXTURE_TS_NSAMPLES)
    return TimeSeries(
        colored_noise / numpy.std(colored_noise) * 5.0,  # Scale noise
        dt=1 / _FIXTURE_TS_SAMPLE_RATE,
        t0=0,
        name="noise",
    )


@pytest.fixture
def gausspulse() -> TimeSeries:
    """0.25s Gaussian pulse at 200 Hz, centred on t0=0."""
    from scipy.signal import gausspulse
    dur = 0.25
    nsamp = int(dur * _FIXTURE_TS_SAMPLE_RATE)
    times = numpy.linspace(-dur/2, dur/2, nsamp, endpoint=False)
    pulse = gausspulse(times, fc=200, bw=.5)
    return TimeSeries(
        pulse,
        times=times,
        name="gausspulse",
    )


@pytest.fixture
def noisy_sinusoid() -> TimeSeries:
    """10s of 2V/rtHz RMS sine wave at 500Hz with 1mV/rtHz white noise (2kHz)."""
    # see :func:`scipy.signal.welch`
    rng = numpy.random.default_rng(seed=1234)

    # params
    freq = 500.
    rate = 2048.
    size = rate * 10

    # create noisy sinusoid
    amp = 2 * numpy.sqrt(2)  # 2V RMS sinusoid
    noise_power = 1e-3 * rate / 2  # mV RMS white noise
    time = numpy.arange(size) / rate
    x = amp * numpy.sin(2 * numpy.pi * freq * time)
    x += rng.normal(scale=numpy.sqrt(noise_power), size=time.shape)
    return TimeSeries(
        x,
        times=time,
        unit="V",
        name="noisy sinusoid",
    )


@pytest.fixture
def corrupt_noisy_sinusoid(noisy_sinusoid: TimeSeries) -> TimeSeries:
    """Return the `noisy_sinusoid` with 10 samples of corruption in the middle."""
    # add corruption in part of the signal
    size = noisy_sinusoid.size
    noisy_sinusoid.value[int(size // 2):int(size // 2) + 10] *= 50.
    noisy_sinusoid.name = "corrupt noisy sinusoid"
    return noisy_sinusoid


# -- NDS2 fixtures -------------------

@pytest.fixture
def nds2_connection(request: pytest.FixtureRequest) -> Iterator[nds2.connection]:
    """Yield an open mocked NDS2 connection.

    Set ``NDS2_CONNECTION_FIXTURE_DATA`` as an iterable of `TimeSeries` in
    the test class or module to customise the data buffers available.

    This fixture also mocks out `os.environ` and ``pop``s any GWDataFind
    server variables, in an attempt to ensure that NDS2 is used by
    `TimeSeries.get` and friends.
    """
    rng = numpy.random.default_rng(seed=0)
    data = None
    for scope in (request.instance, request.module):
        if data := getattr(scope, "NDS2_CONNECTION_FIXTURE_DATA", None):
            break
    else:
        data = [
            TimeSeries(
                rng.random(size=128),
                name="X1:test",
                t0=1000000000,
                sample_rate=16,
                unit="m",
            ),
            TimeSeries(
                rng.random(size=1024),
                name="Y1:test",
                t0=1000000000,
                sample_rate=128,
                unit="V",
            ),
        ]
    nds2conn = mocks.nds2_connection(
        buffers=map(mocks.nds2_buffer_from_timeseries, data),
    )
    mock_environ = mock.patch.dict("os.environ")
    with mock.patch("nds2.connection") as mockconn, mock_environ:
        os.environ.pop("GWDATAFIND_SERVER", "")
        os.environ.pop("LIGO_DATAFIND_SERVER", "")
        mockconn.return_value = nds2conn
        yield nds2conn



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""
.. sectionauthor:: Duncan Macleod <duncan.macleod@ligo.org>
.. currentmodule:: gwpy.timeseries

Comparing seismic trends between LIGO sites
###########################################

On Jan 16 2020 there was a series of earthquakes, that
should have had an impact on LIGO operations, I'd like to find out.
"""

# %%
# Data access
# -----------
# We choose to look at the 0.03Hz-1Hz ground motion band-limited RMS channel
# (1-second average trends) for each interferometer, which records the low-frequency
# motion caused by distant earthquakes and similar seismic events.
# These data are available from the |GWOSC_O3_AUX_TREND_RELEASE|_.
#
# We use a
# `format string <https://docs.python.org/3/tutorial/inputoutput.html#the-string-format-method>`_
# so we can substitute the interferometer prefix without duplicating the channel name:

channel = "{ifo}:ISI-GND_STS_ITMY_Z_BLRMS_30M_100M.rms"
lhochan = channel.format(ifo="H1")
llochan = channel.format(ifo="L1")

# %%
# To access the data, we can use :meth:`~TimeSeriesDict.get`, and give start and end
# datetimes to fetch 6 hours of data for each interferometer:

from gwpy.timeseries import TimeSeriesDict
data = TimeSeriesDict.get(
    [lhochan, llochan],
    "Jan 16 2020 8:00",
    "Jan 16 2020 14:00",
    parallel=8,
)

# %%
# Visualisation
# -------------
# Now that we have the data, we can easily `~TimeSeriesDict.plot` them:

plot = data[lhochan].plot(
    color="gwpy:ligo-hanford",
    label="LIGO-Hanford",
    yscale="log",
    ylabel=r"$1-3$ Hz motion [nm/s]",
)
ax = plot.gca()
ax.plot(data[llochan], color="gwpy:ligo-livingston", label="LIGO-Livingston")
ax.set_title("Impact of earthquakes on LIGO")
ax.legend()
plot.show()

# %%
# As we can see, a series of earthquakes was recorded over this period, potentially
# impacting the LIGO observatories for several hours.
# However, the advanced seismic isolation systems at each site meant that both
# observatories were able to maintain lock throughout this period.



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""HDF5 input output methods to the DataQualityFlag.

While these methods are avialable as methods of the class itself,
this module attaches them to the unified I/O registry, making it a bit
cleaner.
"""

# ruff: noqa: D417

from __future__ import annotations

import warnings

import h5py
import numpy
from astropy.table import Table
from astropy.units import (
    Quantity,
    UnitBase,
)

from ...io import hdf5 as io_hdf5
from ...io.registry import default_registry
from ...time import LIGOTimeGPS
from .. import (
    DataQualityDict,
    DataQualityFlag,
    Segment,
    SegmentList,
)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


# -- read ----------------------------

def _is_flag_group(obj: object) -> bool:
    """Return `True` if `obj` is an `h5py.Group` that looks like a DQ flag."""
    return (
        isinstance(obj, h5py.Group)
        and isinstance(obj.get("active"), h5py.Dataset)
        and isinstance(obj.get("known"), h5py.Dataset)
    )


def _find_flag_groups(h5f: h5py.Group) -> list[str]:
    """Return all groups in `h5f` that look like flags."""
    flag_groups: list[str] = []

    def _find(name: str, obj: object) -> None:
        if _is_flag_group(obj):
            flag_groups.append(name)

    h5f.visititems(_find)
    return flag_groups


def _get_flag_group(
    h5f: h5py.Group,
    path: str | None,
) -> h5py.Group:
    """Determine the group to use in order to read a flag."""
    # if user chose the path, just use it
    if path:
        return h5f[path]

    # if the user gave us the group directly, use it
    if _is_flag_group(h5f):
        return h5f

    # otherwise try and find a single group that matches
    try:
        path, = _find_flag_groups(h5f)
    except ValueError:
        pass
    else:
        return h5f[path]

    # if not exactly 1 valid group in the file, complain
    msg = (
        "please pass a valid HDF5 Group, or specify the HDF5 Group "
        "path via the ``path=`` keyword argument"
    )
    raise ValueError(msg)


@io_hdf5.with_read_hdf5
def read_hdf5_flag(
    h5f: h5py.Group,
    path: str | None = None,
    gpstype: type = LIGOTimeGPS,
) -> DataQualityFlag:
    """Read a `DataQualityFlag` object from an HDF5 file or group."""
    # extract correct group
    dataset = _get_flag_group(h5f, path)

    # read dataset
    active = SegmentList.read(dataset["active"], format="hdf5", gpstype=gpstype)
    try:
        known = SegmentList.read(dataset["known"], format="hdf5", gpstype=gpstype)
    except KeyError as first_keyerror:
        try:
            known = SegmentList.read(dataset["valid"], format="hdf5", gpstype=gpstype)
        except KeyError:
            raise first_keyerror from None

    return DataQualityFlag(active=active, known=known, **dict(dataset.attrs))


@io_hdf5.with_read_hdf5
def read_hdf5_segmentlist(
    h5f: h5py.Group,
    path: str | None = None,
    gpstype: type = LIGOTimeGPS,
    **kwargs,
) -> SegmentList:
    """Read a `SegmentList` object from an HDF5 file or group.

    Parameters
    ----------
    h5f : `str`, `file`, `h5py.Group`, or `list` of
        File path, open file, or HDF5 group from which to read.

    path : `str`, optional
        Path to the dataset inside the HDF5 file/group. If not given, the
        first dataset found will be used.

    gpstype : `type`, optional
        The class to use for the segment endpoints, default is `LIGOTimeGPS`.

    kwargs
        Other keyword arguments passed to :meth:`~astropy.table.Table.read`.
    """
    # find dataset
    dataset = io_hdf5.find_dataset(h5f, path=path)

    # read table
    segtable = Table.read(dataset, format="hdf5", **kwargs)

    # convert to SegmentList
    out = SegmentList()
    for row in segtable:
        start = LIGOTimeGPS(int(row["start_time"]), int(row["start_time_ns"]))
        end = LIGOTimeGPS(int(row["end_time"]), int(row["end_time_ns"]))
        if gpstype is LIGOTimeGPS:
            out.append(Segment(start, end))
        else:
            out.append(Segment(gpstype(start), gpstype(end)))
    return out


@io_hdf5.with_read_hdf5
def read_hdf5_dict(
    h5f: h5py.Group,
    names: list[str] | None = None,
    path: str | None = None,
    on_missing: str = "error",
    **kwargs,
) -> DataQualityDict:
    """Read a `DataQualityDict` from an HDF5 file.

    Parameters
    ----------
    h5f : `str`, `file`, `h5py.Group`, or `list` of
        File path, open file, or HDF5 group (or list of any of these)
        from which to read.

    names : `list` of `str`, optional
        List of flag names to read. If not given, will try and find all
        available names automatically.

    path : `str`, optional
        Path to the group inside the HDF5 file/group in which to search
        for flags. If not given, the whole file/group will be searched.

    on_missing : `str`, optional, default: 'error'
        Action to take if a given name is not found, one of:

        - 'error' : raise an exception
        - 'warn' : issue a warning, but continue
        - 'ignore' : silently skip missing names

    kwargs
        Other keyword arguments are passed to
        `DataQualityFlag.read(..., format="hdf5")` to read each flag.

    Raises
    ------
    ValueError
        If no names are given and automatic detection fails, or if
        ``on_missing='error'`` and a given name is not found.

    Returns
    -------
    dqdict : `DataQualityDict`
        A new `DataQualityDict` of `DataQualityFlag` entries with
        ``active`` and ``known`` segments seeded from the HDF5 groups
        in the given file.
    """
    if path:
        h5f = h5f[path]

    # try and get list of names automatically
    if names is None:
        try:
            names = _find_flag_groups(h5f)
        except KeyError:
            names = None
        if not names:
            msg = (
                "Failed to automatically parse available flag names from "
                "HDF5, please give a list of names to read via the "
                "``names=`` keyword"
            )
            raise ValueError(msg)

    # read data
    out = DataQualityDict()
    for name in names:
        try:
            out[name] = read_hdf5_flag(h5f, name, **kwargs)
        except KeyError as exc:
            if on_missing == "ignore":
                continue
            if on_missing == "warn":
                warnings.warn(str(exc), stacklevel=2)
            else:
                msg = f"no H5Group found for flag '{name}'"
                raise ValueError(msg) from None

    return out


# -- write ---------------------------

def write_hdf5_flag_group(
    flag: DataQualityFlag,
    h5group: h5py.Group,
    **kwargs,
) -> h5py.Group:
    """Write a `DataQualityFlag` into the given HDF5 group."""
    # write segmentlists
    flag.active.write(h5group, "active", **kwargs)
    kwargs["append"] = True
    flag.known.write(h5group, "known", **kwargs)

    # store metadata
    for attr in (
        "name",
        "label",
        "category",
        "description",
        "isgood",
        "padding",
    ):
        value = getattr(flag, attr)
        if value is None:
            continue
        if isinstance(value, Quantity):
            h5group.attrs[attr] = value.value
        elif isinstance(value, UnitBase):
            h5group.attrs[attr] = str(value)
        else:
            h5group.attrs[attr] = value

    return h5group


@io_hdf5.with_write_hdf5
def write_hdf5_dict(
    flags: DataQualityDict,
    output: h5py.Group,
    path: str | None = None,
    **kwargs,
) -> None:
    """Write this `DataQualityFlag` to a `h5py.Group`.

    This allows writing to an HDF5-format file.

    Parameters
    ----------
    output : :class:`h5py.Group`
        Path to new output file, or open h5py `Group` to write to.

    path : `str`, optional
        The HDF5 group path in which to write a new group for this flag.
        If not given, the flag name will be used as the group name.

    append : `bool`, optional
        If `False`, and the target file already exists, an
        `OSError` will be raised. If `True`, the file
        will be opened in append mode. Default is `False`.

    overwrite : `bool`, optional
        If `True`, and the target group already exists, the existing
        group will be deleted before writing the new data. Default is
        `False`.

    kwargs
        Other keyword arguments are passed to `SegmentList.write(format="hdf5")`
        for each segment list.

    Returns
    -------
    dqfgroup : :class:`h5py.Group`
        HDF group containing these data. This group contains 'active'
        and 'known' datasets, and metadata attrs.

    See Also
    --------
    astropy.io
        for details on acceptable keyword arguments when writing a
        :class:`~astropy.table.Table` to HDF5
    """
    if path:
        try:
            parent = output[path]
        except KeyError:
            parent = output.create_group(path)
    else:
        parent = output

    overwrite = kwargs.pop("overwrite", False)

    for name in flags:
        # handle existing group
        if name in parent:
            if not overwrite:
                msg = (
                    f"group '{parent.name}/{name}' already "
                    "exists, give ``overwrite=True`` to "
                    "overwrite it"
                )
                raise OSError(msg)
            del parent[name]
        # create group
        group = parent.create_group(name)
        # write flag
        write_hdf5_flag_group(flags[name], group, **kwargs)


def write_hdf5_flag(
    flag: DataQualityFlag,
    output: h5py.Group,
    path: str | None = None,
    **kwargs,
) -> None:
    """Write a `DataQualityFlag` to an HDF5 file/group.

    Parameters
    ----------
    output : `str`, `h5py.File`, `h5py.Group`, or `list` of these
        Filename or HDF5 object (or list of these) to write to.

    path : `str`, optional
        The HDF5 group path in which to write a new group for this flag.
        If not given, the flag name will be used as the group name.

    kwargs
        Other keyword arguments are passed to `SegmentList.write(format="hdf5")`
        for each segment list.
    """
    # verify path (default to flag name)
    if path is None:
        path = flag.name
    if path is None:
        msg = (
            "cannot determine target group name for flag in HDF5 structure, "
            "please set `name` for each flag, or specify the ``path`` "
            "keyword when writing"
        )
        raise ValueError(msg)

    write_hdf5_dict({path: flag}, output, **kwargs)


def write_hdf5_segmentlist(
    seglist: SegmentList,
    output: h5py.Group,
    path: str | None = None,
    **kwargs,
) -> None:
    """Write a `SegmentList` to an HDF5 file/group.

    Parameters
    ----------
    output : `str`, `h5py.File`, `h5py.Group`
        Filename or HDF5 object to write to.

    path : `str`
        Path to which to write inside the HDF5 file, relative to ``output``.

    kwargs
        Other keyword arguments are passed to
        :meth:`~astropy.table.Table.write`.
    """
    if path is None:
        msg = "please specify the HDF5 path via the ``path=`` keyword argument"
        raise ValueError(msg)

    # convert segmentlist to Table
    data = numpy.zeros((len(seglist), 4), dtype=int)
    for i, seg in enumerate(seglist):
        start, end = map(LIGOTimeGPS, seg)
        data[i, :] = (
            start.gpsSeconds,
            start.gpsNanoSeconds,
            end.gpsSeconds,
            end.gpsNanoSeconds,
        )
    segtable = Table(
        data,
        names=[
            "start_time",
            "start_time_ns",
            "end_time",
            "end_time_ns",
        ],
    )

    # write table to HDF5
    return segtable.write(
        output,
        path=path,
        format="hdf5",
        **kwargs,
    )


# -- register ------------------------

# SegmentList
for klass, read, write in [
    (SegmentList, read_hdf5_segmentlist, write_hdf5_segmentlist),
    (DataQualityFlag, read_hdf5_flag, write_hdf5_flag),
    (DataQualityDict, read_hdf5_dict, write_hdf5_dict),
]:
    default_registry.register_reader("hdf5", klass, read)
    default_registry.register_writer("hdf5", klass, write)
    default_registry.register_identifier("hdf5", klass, io_hdf5.identify_hdf5)



# Copyright (c) 2024-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Core I/O utilities for GWF files."""

from __future__ import annotations

import os
from typing import TYPE_CHECKING

from ...segments import SegmentList
from ..cache import read_cache
from .backend import get_backend_function

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path
    from typing import (
        IO,
        Literal,
    )

    from ...detector import Channel
    from ...io.utils import (
        FileLike,
        FileSystemPath,
    )
    from ...types import Series

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

__all__ = [
    "channel_exists",
    "data_segments",
    "get_channel_names",
    "get_channel_type",
    "identify_gwf",
    "iter_channel_names",
    "num_channels",
]

# first 4 bytes of any valid GWF file (see LIGO-T970130 §4.3.1)
GWF_SIGNATURE = b"IGWD"

FRDATA_TYPES: tuple[str, ...] = (
    "ADC",
    "Proc",
    "Sim",
)


# -- i/o -----------------------------

def identify_gwf(
    origin: Literal["read", "write"],
    filepath: FileSystemPath | None,
    fileobj: FileLike | None,
    *args,
    **kwargs,
) -> bool:
    """Identify a filename or file object as GWF.

    This function is overloaded in that it will also identify a cache file
    as 'gwf' if the first entry in the cache contains a GWF file extension
    """
    # try and read file descriptor
    if fileobj is not None:
        loc = fileobj.tell()
        fileobj.seek(0)
        try:
            signature = fileobj.read(len(GWF_SIGNATURE))
        finally:
            fileobj.seek(loc)
        return signature == GWF_SIGNATURE

    # otherwise read file extension
    if filepath is not None:
        filepath = os.fspath(filepath)
        if filepath.endswith(".gwf"):
            return True
        if filepath.endswith((".lcf", ".cache")):
            try:
                cache = read_cache(filepath)
            except OSError:
                return False
            else:
                return os.fspath(cache[0]).endswith(".gwf")

    return False


# -- utilities -----------------------

def _from_backend(func, backend, *args, **kwargs):
    """Import a function for the given backend and execute it."""
    impl = get_backend_function(func, backend=backend)
    return impl(*args, **kwargs)


def _iter_toc(
    gwf: str | Path | IO,
    type: str | None = None,
    backend: str | None = None,
):
    """Import the ``_iter_toc`` implementation from the backend and run."""
    return _from_backend("_iter_toc", backend, gwf, type=type)


def _count_toc(
    gwf: str | Path | IO,
    type: str | None = None,
    backend: str | None = None,
):
    """Import the ``_count_toc`` implementation from the backend and run."""
    return _from_backend("_count_toc", backend, gwf, type=type)


def num_channels(
    gwf: str | Path | IO,
    type: str | None = None,
    backend: str | None = None,
) -> int:
    """Find the total number of channels in this GWF file.

    Requires a GWF backend library.

    Parameters
    ----------
    gwf : `str`
        GWF file to read.

    type : `str`
        The `Fr` structure type to match (one of 'adc', 'sim', or 'proc').
        Default is to match all types.

    backend : `str`, optional
        The GWF backend to use. Default is 'any'.

    Returns
    -------
    n : `int`
        The total number of channels found in the table of contents.
    """
    return _count_toc(gwf, type=type, backend=backend)


def get_channel_type(
    channel: str | Channel,
    gwf: str | Path | IO,
    backend: str | None = None,
) -> Literal["adc", "sim", "proc"]:
    """Find the channel type in a given GWF file.

    Requires a GWF backend library.

    Parameters
    ----------
    channel : `str`, `~gwpy.detector.Channel`
        Name of data channel to find.

    gwf : `str`
        GWF file to read.

    backend : `str`, optional
        The GWF backend to use. Default is 'any'.

    Returns
    -------
    ctype : `str`
        the type of the channel ('adc', 'sim', or 'proc')

    Raises
    ------
    ValueError
        if the channel is not found in the table-of-contents
    """
    channel = str(channel)
    for name, type_ in _iter_toc(gwf, backend=backend):
        if channel == name:
            return type_
    msg = f"'{channel}' not found in table-of-contents for {gwf}"
    raise ValueError(msg)


def channel_exists(
    gwf: str | Path | IO,
    channel: str | Channel,
    backend: str | None = None,
) -> bool:
    """Determine whether a channel exists in a GWF file.

    Requires a GWF backend library.

    Parameters
    ----------
    channel : `str`
        Name of data channel to find.

    gwf : `str`
        GWF file to read.

    backend : `str`, optional
        The GWF backend to use. Default is 'any'.

    Returns
    -------
    inframe : `bool`
        whether this channel is included in the table of contents for
        the given GWF file.
    """
    try:
        return _from_backend("_channel_exists", backend, gwf, channel)
    except NotImplementedError:
        # no backend-specific optimisation
        return str(channel) in iter_channel_names(gwf, backend=backend)


def iter_channel_names(
    gwf: str | Path | IO,
    type: str | None = None,
    backend: str | None = None,
) -> Iterator[str]:
    """Iterate over the names of channels found in a GWF file.

    Requires a GWF backend library.

    Parameters
    ----------
    gwf : `str`
        GWF file to read.

    type : `str`
        The `Fr` structure type to match (one of 'adc', 'sim', or 'proc').
        Default is to match all types.

    backend : `str`, optional
        The GWF backend to use. Default is 'any'.

    Returns
    -------
    channels : `generator`
        an iterator that will loop over the names of channels as read from
        the table of contents of the given GWF file
    """
    for name, _ in _iter_toc(gwf, type=type, backend=backend):
        yield name


def get_channel_names(
    gwf: str | Path | IO,
    type: str | None = None,
    backend: str | None = None,
) -> list[str]:
    """Return a list of all channel names found in a GWF file.

    This method just returns

    >>> list(iter_channel_names(gwf))

    Requires a GWF backend library.

    Parameters
    ----------
    gwf : `str`
        GWF file to read.

    type : `str`
        The `Fr` structure type to match (one of 'adc', 'sim', or 'proc').
        Default is to match all types.

    backend : `str`, optional
        The GWF backend to use. Default is 'any'.

    Returns
    -------
    channels : `list` of `str`
        a `list` of channel names as read from the table of contents of
        the given GWF file
    """
    return list(iter_channel_names(gwf, type=type, backend=backend))


def data_segments(
    paths: list[str],
    channel: str,
    warn: bool = True,
    backend: str | None = None,
) -> SegmentList:
    """Returns the segments containing data for a channel.

    Requires a GWF backend library.

    A frame is considered to contain data if a valid FrData structure
    (of any type) exists for the channel in that frame.  No checks
    are directly made against the underlying FrVect structures.

    Parameters
    ----------
    paths : `list` of `str`
        A list of GWF file paths.

    channel : `str`
        Name of data channel to find.

    warn : `bool`, optional
        If `True`, emit a `UserWarning` when a channel is not found
        in a frame, otherwise silently ignore.

    backend : `str`, optional
        The GWF backend to use. Default is 'any'.

    Returns
    -------
    segments : `~gwpy.segments.SegmentList`
        the list of segments containing data
    """
    _channel_segments = get_backend_function(
        "_channel_segments",
        backend=backend,
    )
    segments = SegmentList()
    if isinstance(paths, str):
        paths = [paths]
    for path in paths:
        segments.extend(_channel_segments(path, channel, warn=warn))
    return segments.coalesce()


def _series_name(series: Series) -> str | None:
    """Returns the 'name' of a `Series` that should be written to GWF.

    This is basically `series.name or str(series.channel) or ""`.

    Parameters
    ----------
    series : `gwpy.types.Series`
        The input series that will be written.

    Returns
    -------
    name : `str`
        The name to use when storing this series.
    """
    return (
        series.name
        or str(series.channel or "")
        or None
    )



# Copyright (c) 2014-2017 Louisiana State University
#               2018-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for `gwpy.plot.gps`."""

from decimal import Decimal

import numpy
import pytest
from astropy.units import Unit
from matplotlib import pyplot

from ...time import LIGOTimeGPS
from .. import gps as plot_gps


class TestGPSMixin:
    """Tests for `GPSMixin`."""

    TYPE = plot_gps.GPSMixin

    def test_init(self):
        """Test `GPSMixin()`."""
        m = self.TYPE()
        assert m.unit is None
        assert m.epoch is None

    def test_init_args(self):
        """Test `GPSMixin()` with args."""
        m = self.TYPE(unit="second", epoch=100)
        assert m.unit is Unit("second")
        assert m.epoch == 100.

    @pytest.mark.parametrize(("in_", "out"), [
        pytest.param(
            None,
            None,
            id="None",
        ),
        pytest.param(
            1,
            1.,
            id="int",
        ),
        pytest.param(
            "1",
            1.,
            id="str",
        ),
        pytest.param(
            Decimal(12345),
            12345.,
            id="Decimal",
        ),
        pytest.param(
            numpy.float32(56789),
            56789.,
            id="numpy.float",
        ),
        pytest.param(
            LIGOTimeGPS(1234567890, 123000000),
            1234567890.123,
            id="LIGOTimeGPS",
        ),
    ])
    def test_epoch(self, in_, out):
        """Test `GPSMixin.epoch`."""
        mix = self.TYPE(epoch=in_)
        assert mix.epoch == out

    @pytest.mark.parametrize(("in_", "out"), [
        (None, None),
        (Unit("second"), Unit("second")),
        (3600, Unit("hour")),
        ("week", Unit("week")),
        ("weeks", Unit("week")),
    ])
    def test_unit(self, in_, out):
        """Test `GPSMixin.unit`."""
        mix = self.TYPE(unit=in_)
        assert mix.unit == out

    @pytest.mark.parametrize(("badunit", "match"), [
        pytest.param(
            "blah",
            "'blah' did not parse as unit",
            id="bad unit",
        ),
        pytest.param(
            "meter",
            "cannot set GPS unit to 'm'",
            id="wrong type",
        ),
        pytest.param(
            "yoctoday",
            "unrecognised unit 'yoctoday'",
            id="unsupported",
        ),
    ])
    def test_unit_error(self, badunit, match):
        """Test `GPSMixin.unit` error handling."""
        with pytest.raises(
            ValueError,
            match=match,
        ):
            self.TYPE(unit=badunit)

    @pytest.mark.parametrize(("unit", "name"), [
        (None, None),
        ("second", "seconds"),
        (Unit("sday"), "sdays"),  # no long_name
    ])
    def test_get_unit_name(self, unit, name):
        """Test `GPSMixin.get_unit_name()`."""
        mix = self.TYPE(unit=unit)
        assert mix.get_unit_name() == name

    @pytest.mark.parametrize(("unit", "scale"), [
        (None, 1),
        ("second", 1),
        ("minute", 60),
        ("hour", 3600),
    ])
    def test_scale(self, unit, scale):
        """Test `GPSMixin.scale`."""
        mix = self.TYPE(unit=unit)
        assert mix.scale == scale


class TestGpsTransform(TestGPSMixin):
    """Tests for `GPSTransform`."""

    TRANSFORM = plot_gps.GPSTransform
    EPOCH = 100.0
    UNIT = "minutes"
    SCALE = 60.
    X = 190.0
    A = 90.0
    B = 19/6.
    C = 1.5

    def test_init(self):
        """Test `GPSTransform()`."""
        t = self.TRANSFORM()
        assert t.transform(1.0) == 1.0

    def test_epoch(self):
        """Test `GPSTransform.epoch`."""
        transform = self.TRANSFORM(epoch=self.EPOCH)
        assert transform.get_epoch() == self.EPOCH
        assert transform.transform(self.X) == self.A
        assert numpy.isclose(
            transform.inverted().transform(transform.transform(self.X)),
            self.X)

    def test_scale(self):
        """Test `GPSTransform.scale`."""
        transform = self.TRANSFORM(unit=self.UNIT)
        assert transform.get_scale() == self.SCALE
        assert transform.transform(self.X) == self.B
        assert numpy.isclose(
            transform.inverted().transform(transform.transform(self.X)),
            self.X)

    def test_transform(self):
        """Test `GPSTransform.transform`."""
        transform = self.TRANSFORM(epoch=self.EPOCH, unit=self.UNIT)
        assert transform.transform(self.X) == self.C
        assert numpy.isclose(
            transform.inverted().transform(transform.transform(self.X)),
            self.X,
        )


class TestInverseGpsTransform(TestGpsTransform):
    """Tests for `InverseGPSTransform`."""

    TRANSFORM = plot_gps.InvertedGPSTransform
    A = 290.0
    B = 11400.0
    C = 11500.0


@pytest.mark.parametrize(
    "scale",
    sorted(filter(lambda x: x != "auto-gps", plot_gps.GPS_SCALES)),
)
def test_gps_scale(scale):
    """Test `GPSScale` in action."""
    u = Unit(scale[:-1])

    fig = pyplot.figure()
    ax = fig.add_subplot(xscale=scale)
    if scale == "years":
        x = numpy.arange(50)
    else:
        x = numpy.arange(1e2)
    ax.plot(x * u.decompose().scale, x)
    fig.canvas.draw()
    xscale = ax.get_xaxis()._scale
    assert xscale.get_unit() == Unit(scale[:-1])
    pyplot.close(fig)


@pytest.mark.parametrize(("scale", "unit"), [
    (1e-5, "ms"),
    (1e-4, "ms"),
    (1e-3, "s"),
    (1e-2, "s"),
    (1e-1, "s"),
    (1e0, "s"),
    (1e1, "min"),
    (1e2, "min"),
    (1e3, "h"),
    (1e4, "d"),
    (1e5, "wk"),
    (1e6, "wk"),
    (1e7, "yr"),
])
def test_auto_gps_scale(scale, unit):
    """Test ``auto-gps`` scale in action."""
    fig = pyplot.figure()
    ax = fig.add_subplot(xscale="auto-gps")
    ax.plot(numpy.arange(1e2) * scale, numpy.arange(1e2))
    xscale = ax.get_xaxis()._scale
    transform = xscale.get_transform()
    assert transform.unit.name == unit
    pyplot.close(fig)


def test_gps_formatting():
    """Test GPS time formatting."""
    fig = pyplot.figure()
    try:
        ax = fig.gca()
        ax.set_xscale("seconds", epoch=1238040211.67)
        ax.set_xlim(1238040211.17, 1238040212.17)
        fig.canvas.draw()
        ticks = ["-0.5", "-0.4", "-0.3", "-0.2", "-0.1",
                 "0", "0.1", "0.2", "0.3", "0.4", "0.5"]
        assert [x.get_text() for x in ax.get_xticklabels()] == ticks
    finally:
        pyplot.close(fig)



# Copyright (c) 2013-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Registry for FFT averaging methods."""

from __future__ import annotations

from typing import TYPE_CHECKING

from ...utils.decorators import deprecated_function

if TYPE_CHECKING:
    from collections.abc import Callable

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

# registry dict for FFT averaging methods
METHODS: dict[str, Callable] = {}


def _format_name(name: str) -> str:
    return name.lower().replace("-", "_")


def register_method(
    func: Callable,
    name: str | None = None,
    deprecated: bool = False,
) -> str:
    """Register a method of calculating an average spectrogram.

    Parameters
    ----------
    func : `callable`
        function to execute

    name : `str`, optional
        name of the method, defaults to ``func.__name__``

    deprecated : `bool`, optional
        whether this method is deprecated (`True`) or not (`False`)

    Returns
    -------
    name : `str`
        the registered name of the function, which may differ
        pedantically from what was given by the user.
    """
    if name is None:
        name = func.__name__

    # warn about deprecated functions
    if deprecated:
        func = deprecated_function(
            func,
            (
                f"the '{name}' PSD method is deprecated, and will be removed "
                "in a future release, please consider using "
                f"'{name.split('-', 1)[1]}' instead"
            ),
        )

    name = _format_name(name)
    METHODS[name] = func
    return name


def get_method(name: str) -> Callable:
    """Return the PSD method registered with the given name."""
    # find method
    name = _format_name(name)
    try:
        return METHODS[name]
    except KeyError as exc:
        exc.args = f"no PSD method registered with name '{name}'",
        raise



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.table.io.sql`."""

import pytest

from ..io import sql as io_sql


def _sql_str(expr) -> str:
    """Format a SQL expression or string for easy comparison."""
    out = str(expr).lower()
    return out.replace("\n", "")


@pytest.mark.parametrize(("kwargs", "result"), [
    # just table
    pytest.param(
        None,
        'select * from "table"',
        id="table",
    ),
    # columns
    pytest.param(
        {"columns": ["col1", "col2"]},
        'select col1, col2 from "table"',
        id="columns",
    ),
    # 'where' conditions
    pytest.param(
        {
            "columns": ["col1", "col2"],
            "where": "col3 > 4 && col4 = value",
        },
        'select col1, col2 from "table" where col3 > :col3_1 and col4 = :col4_1',
        id="where",
    ),
    # order_by
    pytest.param(
        {
            "order_by": "col5",
        },
        'select * from "table" order by col5 asc',
        id="order_by",
    ),
    # order_by_desc
    pytest.param(
        {
            "columns": ["col1", "col2"],
            "where": ["col3 < 10", "col7 = something"],
            "order_by": "col6",
            "order_by_desc": True,
        },
        'select col1, col2 from "table" where col3 < :col3_1 and col7 = :col7_1 order by col6 desc',  # noqa: E501
        id="order_by_desc",
    ),


])
@pytest.mark.requires("sqlalchemy")
def test_format_query(kwargs, result):
    """Test that `gwpy.table.io.sql.format_query` works."""
    assert _sql_str(io_sql.format_query(
        "table",
        **(kwargs or {}),
    )) == _sql_str(result)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit test for utils module."""

import sys
from math import sqrt

import pytest

from .. import mp as utils_mp

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


@pytest.mark.parametrize("verbose", [False, True, "Test"])
@pytest.mark.parametrize("nproc", [1, 2])
def test_multiprocess_with_queues(capsys, nproc, verbose):
    """Test `multiprocess_with_queues`."""
    inputs = [1, 4, 9, 16, 25]
    out = utils_mp.multiprocess_with_queues(
        nproc,
        sqrt,
        inputs,
        verbose=verbose,
        file=sys.stdout,  # need to pass stdout for capsys to work
    )
    assert out == [1, 2, 3, 4, 5]

    # assert progress bar prints correctly
    stdout = capsys.readouterr().out.strip()
    if verbose is True:
        assert stdout.startswith("Processing: ")
    elif verbose is False:
        assert stdout == ""
    else:
        assert stdout.startswith(verbose)


@pytest.mark.parametrize("nproc", [1, 2])
def test_multiprocess_with_queues_errors(nproc):
    """Check that errors from child processes propagate to the parent."""
    with pytest.raises(
        ValueError,
        match=r"^(math domain error|expected a nonnegative input)",
    ):
        utils_mp.multiprocess_with_queues(
            nproc,
            sqrt,
            [-1, -1, -1, -1],
        )



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.detector.tz`."""

import datetime

import pytest

from .. import tz

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def test_get_timezone():
    """Test `get_timezone()`."""
    assert tz.get_timezone("G1") == "Europe/Berlin"


def test_get_timezone_error():
    """Test `get_timezone()`."""
    with pytest.raises(
        ValueError,
        match=r"^Unrecognised ifo: 'ABC'$",
    ):
        tz.get_timezone("ABC")


@pytest.mark.parametrize(("ifo", "result"), [
    pytest.param("L1", -21600, id="L1"),
    pytest.param("K1", +32400, id="K1"),
])
def test_get_timezone_offset(ifo, result):
    """Test `get_timezone_offset()`."""
    assert tz.get_timezone_offset(
        ifo,
        datetime.datetime(2025, 1, 1, 0, 0, 0, tzinfo=datetime.UTC),
    ) == result


@pytest.mark.freeze_time("2025-1-1 00:00:00")
def test_get_timezone_offset_now():
    """Test `get_timezone_offset()` with default ``dt`` value."""
    assert tz.get_timezone_offset("G1") == 3600



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for interacting with the LIGO Algorithm Library.

This module requires lal >= 6.14.0.
"""

from __future__ import annotations

import operator
import re
from fractions import Fraction
from functools import reduce
from typing import (
    TYPE_CHECKING,
    overload,
)

import lal
import numpy
from astropy import units

# Import gwpy.detector.units to register other units now
from ..detector import units as gwpy_units  # noqa: F401
from ..time import to_gps

if TYPE_CHECKING:
    from collections.abc import Callable
    from types import ModuleType
    from typing import Literal

    from numpy.typing import DTypeLike

    from ..time import SupportsToGps

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

# -- type matching -------------------

#: Mapping of LAL type `int` to type `str`
LAL_TYPE_STR: dict[int, str] = {
    lal.I2_TYPE_CODE: "INT2",
    lal.I4_TYPE_CODE: "INT4",
    lal.I8_TYPE_CODE: "INT8",
    lal.U2_TYPE_CODE: "UINT2",
    lal.U4_TYPE_CODE: "UINT4",
    lal.U8_TYPE_CODE: "UINT8",
    lal.S_TYPE_CODE: "REAL4",
    lal.D_TYPE_CODE: "REAL8",
    lal.C_TYPE_CODE: "COMPLEX8",
    lal.Z_TYPE_CODE: "COMPLEX16",
}

#: Mapping of LAL type `str` to type `int`
LAL_TYPE_FROM_STR: dict[str, int] = {v: k for k, v in LAL_TYPE_STR.items()}

#: Mapping of `numpy` types to LAL type `int`
LAL_TYPE_FROM_NUMPY: dict[type, int] = {
    numpy.int16: lal.I2_TYPE_CODE,
    numpy.int32: lal.I4_TYPE_CODE,
    numpy.int64: lal.I8_TYPE_CODE,
    numpy.uint16: lal.U2_TYPE_CODE,
    numpy.uint32: lal.U4_TYPE_CODE,
    numpy.uint64: lal.U8_TYPE_CODE,
    numpy.float32: lal.S_TYPE_CODE,
    numpy.float64: lal.D_TYPE_CODE,
    numpy.complex64: lal.C_TYPE_CODE,
    numpy.complex128: lal.Z_TYPE_CODE,
}

#: Mapping of `numpy` types to LAL type `str`
LAL_TYPE_STR_FROM_NUMPY: dict[type, str] = {
    k: LAL_TYPE_STR[v] for k, v in LAL_TYPE_FROM_NUMPY.items()
}
#: Mapping of LAL type `str` to `numpy` types
LAL_NUMPY_FROM_TYPE_STR: dict[str, type] = {
    v: k for k, v in LAL_TYPE_STR_FROM_NUMPY.items()
}

#: Regular expression to match LAL type strings
LAL_TYPE_REGEX: re.Pattern[str] = re.compile(r"(U?INT|REAL|COMPLEX)\d+")


def to_lal_type_str(pytype: type | DTypeLike | str | int) -> str:
    """Convert the input python type to a LAL type string.

    Examples
    --------
    To convert a python type:

    >>> from gwpy.utils.lal import to_lal_type_str
    >>> to_lal_type_str(float)
    'REAL8'

    To convert a `numpy.dtype`:

    >>> import numpy
    >>> to_lal_type_str(numpy.dtype('uint32'))
    'UINT4'

    To convert a LAL type code:

    >>> to_lal_type_str(11)
    'REAL8'

    Raises
    ------
    ValueError
        If the input doesn't map to a LAL type string.
    """
    # noop
    if pytype in LAL_TYPE_FROM_STR:
        return pytype  # type: ignore[return-value]

    # convert type code
    if pytype in LAL_TYPE_STR:
        return LAL_TYPE_STR[pytype]  # type: ignore[index]

    # convert python type
    try:
        dtp: type = numpy.dtype(pytype).type  # type: ignore[arg-type]
        return LAL_TYPE_STR_FROM_NUMPY[dtp]
    except (
        TypeError,  # failed to convert input to dtype
        KeyError,  # dtype didn't match
    ) as exc:
        msg = f"Failed to map '{pytype}' to LAL type string"
        raise ValueError(msg) from exc


# -- Type aliases for common types

#: Type alias for LAL FFT plan types
LALFFTPlanType = (
    lal.REAL4FFTPlan
    | lal.REAL8FFTPlan
    | lal.COMPLEX8FFTPlan
    | lal.COMPLEX16FFTPlan
)
#: Type alias for LAL FrequencySeries types
LALFrequencySeriesType = (
    lal.INT2FrequencySeries
    | lal.INT4FrequencySeries
    | lal.INT8FrequencySeries
    | lal.UINT2FrequencySeries
    | lal.UINT4FrequencySeries
    | lal.UINT8FrequencySeries
    | lal.REAL4FrequencySeries
    | lal.REAL8FrequencySeries
    | lal.COMPLEX8FrequencySeries
    | lal.COMPLEX16FrequencySeries
)
#: Type alias for LAL TimeSeries types
LALTimeSeriesType = (
    lal.INT2TimeSeries
    | lal.INT4TimeSeries
    | lal.INT8TimeSeries
    | lal.UINT2TimeSeries
    | lal.UINT4TimeSeries
    | lal.UINT8TimeSeries
    | lal.REAL4TimeSeries
    | lal.REAL8TimeSeries
    | lal.COMPLEX8TimeSeries
    | lal.COMPLEX16TimeSeries
)
#: Type alias for LAL Vector types
LALVectorType = (
    lal.INT2Vector
    | lal.INT4Vector
    | lal.INT8Vector
    | lal.UINT2Vector
    | lal.UINT4Vector
    | lal.UINT8Vector
    | lal.REAL4Vector
    | lal.REAL8Vector
    | lal.COMPLEX8Vector
    | lal.COMPLEX16Vector
)
#: Type alias for LAL Window types
LALWindowType = lal.REAL4Window | lal.REAL8Window


@overload
def find_typed_function(
    pytype: type | DTypeLike,
    prefix: Literal[""],
    suffix: Literal[
        "AverageSpectrumMedian",
        "AverageSpectrumMedianMean",
        "AverageSpectrumWelch",
    ],
    module: ModuleType = lal,
) -> Callable[
    [LALFrequencySeriesType, LALTimeSeriesType, int, int,
     LALWindowType, LALFFTPlanType],
    int,
]: ...

@overload
def find_typed_function(
    pytype: type | DTypeLike,
    prefix: Literal["Create"],
    suffix: Literal["FFTPlan"],
    module: ModuleType = lal,
) -> Callable[[int, int, int], LALFFTPlanType]: ...

@overload
def find_typed_function(
    pytype: type | DTypeLike,
    prefix: Literal["Create"],
    suffix: Literal["FrequencySeries"],
    module: ModuleType = lal,
) -> Callable[
    [str, lal.LIGOTimeGPS, float, float, lal.Unit, int],
    LALFrequencySeriesType,
]: ...

@overload
def find_typed_function(
    pytype: type | DTypeLike,
    prefix: Literal["Create"],
    suffix: Literal["Sequence"],
    module: ModuleType = lal,
) -> Callable[[int], LALVectorType]: ...

@overload
def find_typed_function(
    pytype: type | DTypeLike,
    prefix: Literal["Create"],
    suffix: Literal["TimeSeries"],
    module: ModuleType = lal,
) -> Callable[
    [str | None, lal.LIGOTimeGPS, float, float, lal.Unit, int],
    LALTimeSeriesType,
]: ...

@overload
def find_typed_function(
    pytype: type | DTypeLike,
    prefix: Literal["Create"],
    suffix: Literal["WindowFromSequence"],
    module: ModuleType = lal,
) -> Callable[[LALVectorType], LALWindowType]: ...

@overload
def find_typed_function(
    pytype: type | DTypeLike,
    prefix: Literal["CreateNamed"],
    suffix: Literal["Window"],
    module: ModuleType = lal,
) -> Callable[[str, float, int], LALWindowType]: ...

def find_typed_function(
    pytype: type | DTypeLike,
    prefix: str,
    suffix: str,
    module: ModuleType = lal,
) -> Callable:
    """Return the lal method for the correct type.

    Parameters
    ----------
    pytype : `type`, `numpy.dtype`
        The python type, or dtype, to map.

    prefix : `str`
        The function name prefix (before the type tag).

    suffix : `str`
        The function name suffix (after the type tag).

    module : `ModuleType`
        The module in which to search for the function.
        Default is `lal`.

    Raises
    ------
    AttributeError
        If the function is not found.

    Examples
    --------
    >>> from gwpy.utils.lal import find_typed_function
    >>> find_typed_function(float, 'Create', 'Sequence')
    <built-in function CreateREAL8Sequence>
    """
    laltype = to_lal_type_str(pytype)
    return getattr(module, f"{prefix}{laltype}{suffix}")


def from_lal_type(laltype: type) -> type:
    """Convert the data type of a LAL instance or type into a numpy data type.

    Parameters
    ----------
    laltype : `SwigPyObject` or `type`
        The input LAL instance or type.

    Returns
    -------
    npytype : `type`
        The numpy data type, such as `numpy.uint32`, `numpy.float64`, etc.

    Raises
    ------
    ValueError
        If the numpy data type cannot be inferred from the LAL object.

    Examples
    --------
    >>> from_lal_type(lal.REAL8TimeSeries)
    numpy.float64

    This also works with instances of LAL series types:

    >>> series = lal.CreateINT4TimeSeries("test", 0, 0, 1, "m", 10)
    >>> from_lal_type(series)
    numpy.int32
    """
    if not isinstance(laltype, type):
        laltype = type(laltype)
    name = laltype.__name__
    match = LAL_TYPE_REGEX.match(name)
    if not match or match[0] not in LAL_NUMPY_FROM_TYPE_STR:
        msg = f"{name!r} has no known numpy type equivalent"
        raise ValueError(msg)
    return LAL_NUMPY_FROM_TYPE_STR[match[0]]


# -- units ---------------------------

_LAL_UNIT_INDEX: list[units.UnitBase] = [
    # the order corresponds to how LAL stores compound units
    units.meter,
    units.kilogram,
    units.second,
    units.ampere,
    units.Kelvin,
    units.Unit("strain"),
    units.count,
]


def to_lal_unit(
    astropy_unit: units.Unit | str,
) -> tuple[lal.Unit, float | complex]:
    """Convert the input unit into a |lal.Unit|_ and a scaling factor.

    Parameters
    ----------
    astropy_unit : `~astropy.units.Unit`, `str`
        The input unit.

    Returns
    -------
    lal_unit : |lal.Unit|_
        The LAL representation of the base unit.

    scale : `float`, `complex`
        The linear scaling factor that should be applied to any associated
        data, see _Notes_ below.

    Notes
    -----
    Astropy supports 'scaled' units of the form ``<N> <u>``
    ere ``<N>`` is a `float` and ``<u>`` the base `astropy.units.Unit`,
    e.g. ``'123 m'``, e.g:

    >>> from astropy.units import Quantity
    >>> x = Quantity(4, '123m')
    >>> print(x)
    4.0 123 m
    >>> print(x.decompose())
    492.0 m

    LAL doesn't support scaled units in this way, so this function simply
    returns the scaling factor of the unit so that it may be applied
    manually to any associated data.

    Examples
    --------
    >>> print(to_lal_unit('m**2 / kg ** 4'))
    (m^2 kg^-4, 1.0)
    >>> print(to_lal_unit('123 m'))
    (m, 123.0)

    Raises
    ------
    ValueError
        If LAL doesn't understand the base units for the input.
    """
    # format incoming unit
    aunit = units.Unit(astropy_unit).decompose()

    # handle scaled units
    pow10 = numpy.log10(aunit.scale)
    if pow10 and pow10.is_integer():
        lunit = lal.Unit(f"10^{int(pow10)}")
        scale = 1
    else:
        lunit = lal.Unit()
        scale = aunit.scale

    # decompose unit into LAL base units
    for base, power in zip(aunit.bases, aunit.powers, strict=True):
        try:  # try this base
            i = _LAL_UNIT_INDEX.index(base)
        except ValueError as exc:
            exc.args = (f"LAL has no unit corresponding to '{base}'",)
            raise
        frac = Fraction(power)
        lunit.unitNumerator[i] = frac.numerator
        lunit.unitDenominatorMinusOne[i] = frac.denominator - 1

    return lunit, scale


def from_lal_unit(
    lunit: lal.Unit,
) -> units.Unit:
    """Convert a LALUnit` into a `~astropy.units.Unit`.

    Parameters
    ----------
    lunit : |lal.Unit|_
        The input unit.

    Returns
    -------
    unit : `~astropy.units.Unit`
        The Astropy representation of the input.

    Raises
    ------
    TypeError
        If ``lunit`` cannot be converted to |lal.Unit|_.

    ValueError
        If Astropy doesn't understand the base units for the input.
    """
    return reduce(
        operator.mul,
        (
            _LAL_UNIT_INDEX[i] ** Fraction(int(num), int(den + 1))
            for i, (num, den) in enumerate(zip(
                lunit.unitNumerator,
                lunit.unitDenominatorMinusOne,
                strict=True,
            ))
        ),
    ) * 10 ** lunit.powerOfTen


def to_lal_ligotimegps(
    gps: SupportsToGps,
) -> lal.LIGOTimeGPS:
    """Convert the given GPS time to a |lal.LIGOTimeGPS|_ object.

    Parameters
    ----------
    gps : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        Input GPS time, can be anything parsable by :meth:`~gwpy.time.to_gps`.

    Returns
    -------
    ligotimegps : |lal.LIGOTimeGPS|_
        A SWIG-LAL `~lal.LIGOTimeGPS` representation of the given GPS time.
    """
    gps = to_gps(gps)
    return lal.LIGOTimeGPS(gps.gpsSeconds, gps.gpsNanoSeconds)


# -- detectors -----------------------

#: Mapping of IFO names to |lal.FrDetector|_ objects
LAL_DETECTORS: dict[str, lal.FrDetector] = {
    ifo.frDetector.prefix: ifo.frDetector
    for ifo in lal.CachedDetectors
}



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.io.nds2`."""

from __future__ import annotations

import os
import warnings
from typing import (
    Generic,
    TypeVar,
)
from unittest import mock

import pytest

from ...detector import Channel
from ...segments import (
    Segment,
    SegmentList,
)
from ...testing import (
    mocks,
    utils,
)
from ...time import LIGOTimeGPS
from ...utils.tests.test_enum import TestNumpyTypeEnum as _TestNumpyTypeEnum
from .. import nds2 as io_nds2

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

EnumType = TypeVar("EnumType", bound=io_nds2._Nds2Enum)


class _TestNds2Enum(Generic[EnumType]):
    """Base class for tests of :class:`gwpy.io.nds2._Nds2Enum`."""

    TEST_CLASS: type[EnumType]

    def test_any(self):
        assert self.TEST_CLASS.any() == 2 * max(self.TEST_CLASS).value - 1

    def test_find_unknown(self):
        """Test 'UNKNOWN' for :meth:`gwpy.io.nds2.Nds2ChannelType.find`."""
        assert self.TEST_CLASS.find("blah").name == "UNKNOWN"


class TestNds2ChannelType(_TestNds2Enum[io_nds2.Nds2ChannelType]):
    """Tests of :class:`gwpy.io.nds2.Nds2DataType`."""

    TEST_CLASS: type[io_nds2.Nds2ChannelType] = io_nds2.Nds2ChannelType

    def test_nds2name(self):
        """Test the `nds2name` property."""
        assert self.TEST_CLASS.MTREND.nds2name == "m-trend"

    def test_nds2names(self):
        """Test the `nds2names` method."""
        expected = sorted(x.nds2name for x in self.TEST_CLASS)
        assert sorted(self.TEST_CLASS.nds2names()) == expected

    @pytest.mark.parametrize(("input_", "expected"), [
        (TEST_CLASS.MTREND.value, TEST_CLASS.MTREND),
        (TEST_CLASS.MTREND.name, TEST_CLASS.MTREND),
        (TEST_CLASS.MTREND.nds2name, TEST_CLASS.MTREND),
        ("mtrend", TEST_CLASS.MTREND),
        ("rds", TEST_CLASS.RDS),
        ("RDS", TEST_CLASS.RDS),
        ("reduced", TEST_CLASS.RDS),
        ("REDUCED", TEST_CLASS.RDS),
    ])
    def test_find(self, input_, expected):
        """Test :meth:`gwpy.io.nds2.Nds2ChannelType.find`."""
        assert self.TEST_CLASS.find(input_) == expected


class TestNds2DataType(
    _TestNds2Enum[io_nds2.Nds2DataType],
    _TestNumpyTypeEnum,
):
    """Tests of :class:`gwpy.io.nds2.Nds2DataType`."""

    TEST_CLASS = io_nds2.Nds2DataType  # type: ignore[assignment]

    def test_find_errors(self):
        """Test errors from :meth:`gwpy.io.nds2.Nds2DataType.find`."""
        pytest.skip(f"not implemented for {self.TEST_CLASS.__name__}")


@pytest.mark.parametrize(("key", "value", "result"), [
    pytest.param(
        "NDSSERVER",
        "test1.ligo.org:80,test2.ligo.org:43",
        [("test1.ligo.org", 80), ("test2.ligo.org", 43)],
        id="basic",
    ),
    pytest.param(
        "NDSSERVER",
        "test1.ligo.org:80,test2.ligo.org:43,test.ligo.org,test2.ligo.org:43",
        [("test1.ligo.org", 80), ("test2.ligo.org", 43), ("test.ligo.org", None)],
        id="repeat",
    ),
    pytest.param(
        "TESTENV",
        "test1.ligo.org:80,test2.ligo.org:43",
        [("test1.ligo.org", 80), ("test2.ligo.org", 43)],
        id="env-key",
    ),
])
def test_parse_nds_env(key, value, result):
    """Test `gwpy.io.nds2.parse_nds_env`."""
    with mock.patch.dict(os.environ, {key: value}):
        if key == "NDSSERVER":
            hosts = io_nds2.parse_nds_env()
        else:
            hosts = io_nds2.parse_nds_env(env=key)
    assert hosts == result


@mock.patch.dict(os.environ, clear=True)
@pytest.mark.parametrize(("ifo", "include_gwosc", "result"), [
    pytest.param(
        None,
        True,
        [("nds.ligo.caltech.edu", None), ("nds.gwosc.org", None)],
        id="default",
    ),
    pytest.param(
        "L1",
        False,
        [("nds.ligo-la.caltech.edu", None), ("nds.ligo.caltech.edu", None)],
        id="L1",
    ),
])
def test_host_resolution_order(ifo, include_gwosc, result):
    """Test `gwpy.io.nds2.host_resolution_order` basic usage."""
    assert io_nds2.host_resolution_order(
        ifo,
        include_gwosc=include_gwosc,
    ) == result


@mock.patch.dict(
    os.environ,
    {"NDSSERVER": "test1.ligo.org:80,test2.ligo.org:43"},
)
@pytest.mark.parametrize(("ifo", "include_gwosc", "result"), [
    pytest.param(
        None,
        False,
        [
            ("test1.ligo.org", 80),
            ("test2.ligo.org", 43),
            ("nds.ligo.caltech.edu", None),
        ],
        id="noifo",
    ),
    pytest.param(
        "L1",
        True,
        [
            ("test1.ligo.org", 80),
            ("test2.ligo.org", 43),
            ("nds.ligo-la.caltech.edu", None),
            ("nds.ligo.caltech.edu", None),
            ("nds.gwosc.org", None),
        ],
        id="L1",
    ),
])
def test_host_resolution_order_env(ifo, include_gwosc, result):
    """Test `host_resolution_order` environment parsing."""
    assert io_nds2.host_resolution_order(
        ifo,
        include_gwosc=include_gwosc,
    ) == result


@mock.patch.dict(
    os.environ,
    {"TESTENV": "test1.ligo.org:80,test2.ligo.org:43"},
)
def test_host_resolution_order_named_env():
    """Test `host_resolution_order` with a named environment variable."""
    hro = io_nds2.host_resolution_order(
        "V1",
        env="TESTENV",
        include_gwosc=False,
    )
    assert hro == [
        ("test1.ligo.org", 80),
        ("test2.ligo.org", 43),
        ("nds.ligo.caltech.edu", None),
    ]


@mock.patch.dict(
    os.environ,
    {"TESTENV": "test1.ligo.org:80,test2.ligo.org:43"},
)
@pytest.mark.parametrize(("ifo", "epoch", "env", "result"), [
    pytest.param(
        "L1",
        "Jan 1 2015",
        None,
        [
            ("nds.ligo.caltech.edu", None),
            ("nds.ligo-la.caltech.edu", None),
        ],
        id="old",
    ),
    pytest.param(
        "L1",
        "now",
        "TESTENV",
        [
            ("test1.ligo.org", 80),
            ("test2.ligo.org", 43),
            ("nds.ligo-la.caltech.edu", None),
            ("nds.ligo.caltech.edu", None),
        ],
        id="now",
    ),
])
def test_host_resolution_order_epoch(ifo, epoch, env, result):
    """Test `gwpy.io.nds2.host_resolution_order` epoch parsing."""
    assert io_nds2.host_resolution_order(
        ifo,
        epoch=epoch,
        env=env,
        include_gwosc=False,
    ) == result


@mock.patch.dict(
    os.environ,
    {"TESTENV": "test1.ligo.org:80,test2.ligo.org:43"},
)
def test_host_resolution_order_warning():
    """Test `gwpy.io.nds2.host_resolution_order` warnings."""
    # test warnings for unknown IFO
    with pytest.warns(
        UserWarning,
        match="no default host found for ifo 'X1'",
    ):
        # should produce warning
        hro = io_nds2.host_resolution_order(
            "X1",
            env=None,
            include_gwosc=False,
        )
    assert hro == [("nds.ligo.caltech.edu", None)]

    with warnings.catch_warnings():
        warnings.simplefilter("error")
        # should _not_ produce warning
        hro = io_nds2.host_resolution_order("X1", env="TESTENV")


@pytest.mark.requires("nds2")
@pytest.mark.parametrize(("host", "port", "callport"), [
    pytest.param("nds.test.gwpy", None, None, id="None"),
    pytest.param("nds.test.gwpy", 31200, 31200, id="31200"),
    pytest.param("x1nds9", None, 8088, id="nds1"),
])
@mock.patch("nds2.connection")
def test_connect(connector, host, port, callport):
    """Test `gwpy.io.nds2.connect`."""
    io_nds2.connect(host, port=port)
    if callport is None:
        connector.assert_called_once_with(host)
    else:
        connector.assert_called_once_with(host, callport)


@pytest.mark.requires("nds2")
@mock.patch("gwpy.io.nds2.connect")
def test_auth_connect(connect):
    """Test `gwpy.io.nds2.auth_connect`."""
    io_nds2.auth_connect("host", 0)
    connect.assert_called_once_with("host", 0)


@pytest.mark.requires("nds2")
@mock.patch("gwpy.io.nds2.kinit")
@mock.patch(
    "gwpy.io.nds2.connect",
    side_effect=(
        RuntimeError("Request SASL authentication something something"),
        True,
    ),
)
def test_auth_connect_kinit(connect, kinit, caplog):
    """Test `gwpy.io.nds2.auth_connect` with a callout to `gwpy.io.kerberos.kinit`."""
    with caplog.at_level("WARNING"):
        assert io_nds2.auth_connect("host", 0)
    assert "attempting Kerberos kinit()" in caplog.text
    kinit.assert_called_with()
    assert connect.call_count == 2
    connect.assert_called_with("host", 0)


@pytest.mark.requires("nds2")
@mock.patch(
    "gwpy.io.nds2.connect",
    side_effect=RuntimeError("Anything else"),
)
def test_auth_connect_error(connect):
    """Test errors from `gwpy.io.nds2.auth_connect`."""
    with pytest.raises(
        RuntimeError,
        match="Anything else",
    ):
        io_nds2.auth_connect("host", 0)
    connect.assert_called_once_with("host", 0)


@pytest.mark.requires("nds2")
def test_find_channels(nds2_connection):
    """Test `gwpy.io.nds2.find_channels`."""
    # call function and check result
    chans = io_nds2.find_channels(
        ["X1:test"],
        host="test.nds2",
    )
    assert len(chans) == 1
    assert chans[0].name == "X1:test"

    # check callouts in find_channels
    nds2_connection.set_epoch.assert_any_call("ALL")

    # check callouts in _find_channels
    nds2_connection.find_channels.assert_called_once_with(
        "X1:test",
        io_nds2.Nds2ChannelType.any(),
        io_nds2.Nds2DataType.any(),
        io_nds2.MIN_SAMPLE_RATE,
        io_nds2.MAX_SAMPLE_RATE,
    )


@pytest.mark.requires("nds2")
def test_find_channels_nds1(nds2_connection):
    """Test NDS1 name handling in `find_channels`."""
    nds2_connection.get_protocol.return_value = 1
    io_nds2.find_channels(
        ["X1:test,m-trend"],
        host="test.nds2",
        sample_rate=16,
        dtype=float,
    )
    nds2_connection.find_channels.assert_called_with(
        "X1:test,m-trend",
        io_nds2.Nds2ChannelType.MTREND.value,
        io_nds2.Nds2DataType.FLOAT64.value,
        16,
        16,
    )


@pytest.mark.requires("nds2")
def test_find_channels_online(nds2_connection):
    """Test handling of 'online' channels in `find_channels`."""
    # add an 'online' version of thee same channel
    buff = nds2_connection._buffers[0]
    nds2_connection._buffers.append(
        mocks.nds2_buffer(
            buff.name,
            buff.data,
            LIGOTimeGPS(buff.gps_seconds, buff.gps_nanoseconds),
            buff.sample_rate,
            buff.channel.signal_units,
            channel_type=int(io_nds2.Nds2ChannelType.ONLINE),
        ),
    )
    assert io_nds2.find_channels(
        ["X1:test"],
        host="test",
        unique=True,
    ) == [nds2_connection._buffers[0].channel]


@pytest.mark.requires("nds2")
def test_find_channels_unique(nds2_connection):
    """Test handling of 'online' channels in `find_channels`."""
    # add a second copy of the same channel
    # so that find_channels returns two things
    nds2_connection._buffers.append(nds2_connection._buffers[0])

    with pytest.raises(
        ValueError,
        match=r"^unique NDS2 channel match not found for 'X1:test'$",
    ):
        io_nds2.find_channels(["X1:test"], host="test", unique=True)


@pytest.mark.requires("nds2")
def test_get_availability(nds2_connection):
    """Test `gwpy.io.nds2.get_availability`."""
    # validate call and parsing of results
    segs = io_nds2.get_availability(
        ["X1:test"],
        0,
        1,
        host="test",
    )
    utils.assert_dict_equal(
        segs,
        {"X1:test": SegmentList([Segment(1000000000.0, 1000000008.0)])},
        utils.assert_segmentlist_equal,
    )

    # check callouts
    nds2_connection.set_epoch.assert_has_calls([
        mock.call(0, 1),
        mock.call(
            nds2_connection.current_epoch().gps_start,
            nds2_connection.current_epoch().gps_stop,
        ),
    ])
    nds2_connection.get_availability.assert_called_once_with(["X1:test,raw"])


@pytest.mark.requires("nds2")
def test_get_availability_real():
    """Test `gwpy.io.nds2.get_availability` with real remote data."""
    try:
        segs = io_nds2.get_availability(
            ["L1:ISI-GND_STS_ITMY_Z_BLRMS_30M_100M"],
            1238166018,
            1238192569,
            host="nds.gwosc.org",
        )
    except RuntimeError as exc:  # something went wrong with NDS2
        pytest.skip(str(exc))
    if sum(len(seglist) for seglist in segs.values()) == 0:  # EMPTY response
        pytest.skip("no data received from nds.gwosc.org")
    utils.assert_dict_equal(
        segs,
        {"L1:ISI-GND_STS_ITMY_Z_BLRMS_30M_100M": SegmentList([
            Segment(1238166018, 1238170289),
            Segment(1238175433, 1238192569),
        ])},
        utils.assert_segmentlist_equal,
    )


@pytest.mark.parametrize(("start", "end", "out"), [
    pytest.param(
        0,
        60,
        (0, 60),
        id="no change",
    ),
    pytest.param(
        1,
        60,
        (0, 60),
        id="expand start",
    ),
    pytest.param(
        0,
        61,
        (0, 120),
        id="expand end",
    ),
    pytest.param(
        59,
        61,
        (0, 120),
        id="expand both",
    ),
    pytest.param(
        1167264018,
        1198800018,
        (1167264000, 1198800060),
        id="expand both GPS",
    ),
])
def test_minute_trend_times(start, end, out):
    """Test `gwpy.io.nds2.minute_trend_times`."""
    assert io_nds2.minute_trend_times(start, end) == out


@pytest.mark.requires("nds2")
def test_get_nds2_name():
    """Test `gwpy.io.nds2._get_nds2_name`."""
    # Note: we can't use parametrize because mocks.nds2_channel requires
    #       the nds2-client and is executed _before_ the skip decorator is
    #       applied
    for channel, name in [
        ("test", "test"),
        (Channel("X1:TEST", type="m-trend"), "X1:TEST,m-trend"),
        (mocks.nds2_channel("X1:TEST", 16, "NONE"), "X1:TEST,raw"),
    ]:
        assert io_nds2._get_nds2_name(channel) == name  # type: ignore[arg-type]


@pytest.mark.requires("nds2")
def test_get_nds2_names():
    """Test `gwpy.io.nds2._get_nds2_names`."""
    channels = [
        "test",
        Channel("X1:TEST", type="m-trend"),
        mocks.nds2_channel("X1:TEST", 16, "NONE"),
    ]
    names = [
        "test",
        "X1:TEST,m-trend",
        "X1:TEST,raw",
    ]
    assert list(io_nds2._get_nds2_names(channels)) == names  # type: ignore[arg-type]



# Copyright (c) 2016-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Python implementation of the tiled Q-transform scan.

This is a re-implementation of the original Q-transform scan from the Omega
pipeline, all credits for the original algorithm go to its
authors.
"""

from __future__ import annotations

import warnings
from math import (
    ceil,
    exp,
    isinf,
    log,
    pi,
)
from typing import TYPE_CHECKING

import numpy
from numpy import fft as npfft

from ..segments import Segment
from ..utils import round_to_power

if TYPE_CHECKING:
    from collections.abc import Iterator

    from numpy.typing import NDArray

    from ..frequencyseries import FrequencySeries
    from ..spectrogram import Spectrogram
    from ..table import EventTable
    from ..time import LIGOTimeGPS
    from ..timeseries import TimeSeries

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__credits__ = [
    "Scott Coughlin <scott.coughlin@ligo.org>, ",
    "Alex Urban <alexander.urban@ligo.org>",
]
__all__ = [
    "QGram",
    "QPlane",
    "QTile",
    "QTiling",
    "q_scan",
]

# -- q-transform defaults

#: Default ``(fmin, fmax)`` frequency range for Q-transform.
DEFAULT_FRANGE = (0, float("inf"))

#: Default maximum fractional mismatch between neighbouring tiles.
DEFAULT_MISMATCH = 0.2

#: Default ``(qlow, qhigh)`` Q range for Q-transform.
DEFAULT_QRANGE = (4, 64)

# constants
_SQRT_2 = 2 ** (1 / 2.)


# -- object class definitions --------

class QObject:
    """Base class for Q-transform objects.

    This object exists just to provide basic methods for all other
    Q-transform objects.
    """

    def __init__(
        self,
        duration: float,
        sampling: float,
        mismatch: float = DEFAULT_MISMATCH,
    ) -> None:
        """Initialize a `QObject`."""
        self.duration = float(duration)
        self.sampling = float(sampling)
        self.mismatch = float(mismatch)

    @property
    def deltam(self) -> float:
        """Fractional mismatch between neighbouring tiles."""
        return 2 * (self.mismatch / 3.) ** (1 / 2.)


class QBase(QObject):
    """Base class for Q-transform objects with fixed Q.

    This class just provides a property for Q-prime = Q / sqrt(11)
    """

    def __init__(
        self,
        q: float,
        duration: float,
        sampling: float,
        mismatch: float = DEFAULT_MISMATCH,
    ) -> None:
        """Initialize a `QBase`."""
        super().__init__(
            duration,
            sampling,
            mismatch=mismatch,
        )
        self.q = float(q)

    @property
    def qprime(self) -> float:
        """Normalized Q `(q/sqrt(11))`."""
        return self.q / 11 ** (1 / 2.)


class QTiling(QObject):
    """Iterable constructor of `QPlane` objects.

    For a given Q-range, each of the resulting `QPlane` objects can
    be iterated over.

    Parameters
    ----------
    duration : `float`
        The duration of the data to be Q-transformed.

    qrange : `tuple` of `float`
        ``(low, high)`` pair of Q extrema.

    frange : `tuple` of `float`
        ``(low, high)`` pair of frequency extrema.

    sampling : `float`
        Sampling rate (in Hertz) of data to be Q-transformed.

    mismatch : `float`
        Maximum fractional mismatch between neighbouring tiles.
    """

    def __init__(
        self,
        duration: float,
        sampling: float,
        qrange: tuple[float, float] = DEFAULT_QRANGE,
        frange: tuple[float, float] = DEFAULT_FRANGE,
        mismatch: float = DEFAULT_MISMATCH,
    ) -> None:
        """Initialize a `QTiling` object."""
        super().__init__(
            duration,
            sampling,
            mismatch=mismatch,
        )
        self.qrange = (float(qrange[0]), float(qrange[1]))

        # format (min, max) frequency range
        qlist = list(self._iter_qs())
        fmin, fmax = float(frange[0]), float(frange[1])
        if fmin == 0:  # set non-zero lower frequency
            fmin = 50 * max(qlist) / (2 * pi * self.duration)
        maxf = (
            self.sampling
            / 2
            / (1 + 11 ** (1 / 2.) / min(qlist))
        )
        if isinf(fmax):
            fmax = maxf
        elif fmax > maxf:  # truncate upper frequency to maximum
            warnings.warn(
                f"upper frequency of {fmax} Hz is too high for the given "
                f"Q range, resetting to {maxf} Hz",
                stacklevel=2,
            )
            fmax = maxf
        self.frange = (fmin, fmax)

    @property
    def qs(self) -> numpy.ndarray:
        """Array of Q values for this `QTiling`."""
        return numpy.array(list(self._iter_qs()))

    @property
    def whitening_duration(self) -> float:
        """The recommended data duration required for whitening."""
        return max(t.whitening_duration for t in self)

    def _iter_qs(self) -> Iterator[float]:
        """Iterate over the Q values."""
        # work out how many Qs we need
        sqrt_2 = _SQRT_2
        cumum = log(self.qrange[1] / self.qrange[0]) / sqrt_2
        nplanes = int(max(ceil(cumum / self.deltam), 1))
        delta_q = cumum / nplanes
        for i in range(nplanes):
            yield self.qrange[0] * exp(sqrt_2 * delta_q * (i + .5))

    def __iter__(self) -> Iterator[QPlane]:
        """Iterate over this `QTiling`.

        Yields a `QPlane` at each Q value.
        """
        for q in self._iter_qs():
            yield QPlane(
                q,
                self.frange,
                self.duration,
                self.sampling,
                mismatch=self.mismatch,
            )

    def transform(
        self,
        fseries: FrequencySeries,
        **kwargs,
    ) -> tuple[QGram, float]:
        """Compute the time-frequency plane at fixed Q with the most significant tile.

        Parameters
        ----------
        fseries : `~gwpy.timeseries.FrequencySeries`
            The complex FFT of a time-series data set.

        kwargs
            Other keyword arguments to pass to `QPlane.transform`.

        Returns
        -------
        out : `QGram`
            Signal energies over the time-frequency plane containing the most
            significant tile.

        far : `float`
            Crudely estimated false alarm rate (FAR) for the detected signal.

        See Also
        --------
        QPlane.transform
            Compute the Q-transform over a single time-frequency plane.
        """
        if not numpy.isfinite(fseries).all():
            msg = "Input signal contains non-numerical values"
            raise ValueError(msg)
        weight: float = 1 + numpy.log10(
            self.qrange[1] / self.qrange[0],
        ) / numpy.sqrt(2)
        nind, nplanes, peak, result = (0., 0, 0., None)
        # identify the plane with the loudest tile
        for plane in self:
            nplanes += 1
            nind += sum([1 + row.ntiles * row.deltam for row in plane])
            result = plane.transform(fseries, **kwargs)
            if result.peak["energy"] > peak:
                out = result
                peak = out.peak["energy"]

        ntiles = nind * weight / nplanes
        far = 1.5 * ntiles * numpy.exp(-out.peak["energy"]) / self.duration
        return out, far


class QPlane(QBase):
    """Iterable representation of a Q-transform plane.

    For a given Q, an array of frequencies can be iterated over, yielding
    a `QTile` each time.

    Parameters
    ----------
    q : `float`
        The Q-value for this plane.

    frange : `tuple` of `float`
        ``(low, high)`` range of frequencies for this plane.

    duration : `float`
        The duration of the data to be Q-transformed.

    sampling : `float`
        Sampling rate (in Hertz) of data to be Q-transformed.

    mismatch : `float`
        Maximum fractional mismatch between neighbouring tiles.
    """

    def __init__(
        self,
        q: float,
        frange: tuple[float, float],
        duration: float,
        sampling: float,
        mismatch: float = DEFAULT_MISMATCH,
    ) -> None:
        """Initialize a `QPlane` object."""
        super().__init__(
            q,
            duration,
            sampling,
            mismatch=mismatch,
        )

        # format (min, max) frequency range
        fmin, fmax = float(frange[0]), float(frange[1])
        if fmin == 0:  # set non-zero lower frequency
            fmin = 50 * self.q / (2 * pi * self.duration)
        if isinf(fmax):  # set non-infinite upper frequency
            fmax = self.sampling / 2 / (1 + 1 / self.qprime)
        self.frange = (fmin, fmax)

    def __iter__(self) -> Iterator[QTile]:
        """Iterate over this `QPlane`.

        Yields a `QTile` at each frequency.
        """
        # for each frequency, yield a QTile
        for freq in self._iter_frequencies():
            yield QTile(
                self.q,
                freq,
                self.duration,
                self.sampling,
                mismatch=self.mismatch,
            )

    def _iter_frequencies(self) -> Iterator[float]:
        """Iterate over the frequencies of this `QPlane`."""
        # work out how many frequencies we need
        minf, maxf = self.frange
        fcum_mismatch = log(maxf / minf) * (2 + self.q ** 2) ** (1 / 2.) / 2.
        nfreq = int(max(1, ceil(fcum_mismatch / self.deltam)))
        fstep = fcum_mismatch / nfreq
        fstepmin = 1 / self.duration
        # for each frequency, yield a QTile
        last = None
        for i in range(nfreq):
            this = (
                minf
                * exp(2 / (2 + self.q ** 2) ** (1 / 2.) * (i + .5) * fstep)
                // fstepmin * fstepmin
            )
            if this != last:  # yield only unique elements
                yield this
            last = this

    @property
    def frequencies(self) -> NDArray[numpy.float64]:
        """Array of central frequencies for this `QPlane`."""
        return numpy.array(list(self._iter_frequencies()))

    @property
    def farray(self) -> NDArray[numpy.float64]:
        """Array of frequencies for the lower-edge of each frequency bin."""
        bandwidths = 2 * pi ** (1 / 2.) * self.frequencies / self.q
        return self.frequencies - bandwidths / 2.

    @property
    def whitening_duration(self) -> float:
        """The recommended data duration required for whitening."""
        return round_to_power(
            self.q / (2 * self.frange[0]),
            base=2,
            which=None,
        )

    def transform(
        self,
        fseries: FrequencySeries,
        *,
        norm: bool | str = True,
        epoch: float | LIGOTimeGPS | None = None,
        search: Segment | None = None,
    ) -> QGram:
        """Calculate the energy `QGram` for the given `fseries`.

        Parameters
        ----------
        fseries : `~gwpy.frequencyseries.FrequencySeries`
            The complex FFT of a time-series data set.

        norm : `bool`, `str`
            Normalize the energy of the output by the median (if `True` or
            ``'median'``) or the ``'mean'``, if `False` the output
            is the complex `~numpy.fft.ifft` output of the Q-tranform.

        epoch : `~gwpy.time.LIGOTimeGPS`, `float`
            The epoch of these data, only used for metadata in the output
            `TimeSeries`, and not requires if the input `fseries` has the
            epoch populated.

        search : `~gwpy.segments.Segment`
            Search window of interest to determine the loudest Q-plane.

        Returns
        -------
        results : `QGram`
            The complex energies of the Q-transform of the input `fseries`
            at each frequency.

        See Also
        --------
        QTile.transform
            For details on the transform over a row of `(Q, frequency)` tiles.
        QGram
            An object with energies populated over time-frequency tiles.
        """
        out = [qtile.transform(
            fseries,
            norm=norm,
            epoch=epoch,
        ) for qtile in self]
        return QGram(self, out, search)


class QTile(QBase):
    """Representation of a tile with fixed Q and frequency."""

    def __init__(
        self,
        q: float,
        frequency: float,
        duration: float,
        sampling: float,
        mismatch: float = DEFAULT_MISMATCH,
    ) -> None:
        """Initialize a `QTile` object."""
        super().__init__(
            q,
            duration,
            sampling,
            mismatch=mismatch,
        )
        self.frequency = frequency

    @property
    def bandwidth(self) -> float:
        """The bandwidth for tiles in this row."""
        return 2 * pi ** (1 / 2.) * self.frequency / self.q

    @property
    def ntiles(self) -> int:
        """The number of tiles in this row."""
        tcum_mismatch = self.duration * 2 * pi * self.frequency / self.q
        return int(round_to_power(
            tcum_mismatch / self.deltam,
            base=2,
            which="upper",
        ))

    @property
    def windowsize(self) -> int:
        """The size of the frequency-domain window for this row."""
        return 2 * int(
            self.frequency
            / self.qprime
            * self.duration,
        ) + 1

    def _get_indices(self) -> NDArray[numpy.int64]:
        half = int((self.windowsize - 1) / 2)
        return numpy.arange(-half, half + 1)

    def get_window(self) -> NDArray[numpy.float64]:
        """Generate the bi-square window for this row.

        Returns
        -------
        window : `numpy.ndarray`
        """
        # real frequencies
        wfrequencies = self._get_indices() / self.duration
        # dimensionless frequencies
        xfrequencies = wfrequencies * self.qprime / self.frequency
        # normalize and generate bi-square window
        norm = (
            self.ntiles
            / (self.duration * self.sampling)
            * (315 * self.qprime / (128 * self.frequency)) ** (1 / 2.)
        )
        return (1 - xfrequencies ** 2) ** 2 * norm

    def get_data_indices(self) -> numpy.ndarray[tuple[int], numpy.dtype[numpy.int64]]:
        """Return the index array of interesting frequencies for this row."""
        return numpy.round(
            self._get_indices() + 1 + self.frequency * self.duration,
        ).astype(int)

    @property
    def padding(self) -> tuple[int, int]:
        """The `(left, right)` padding required for the IFFT."""
        pad = self.ntiles - self.windowsize
        return (int((pad - 1) / 2.), int((pad + 1) / 2.))

    def transform(
        self,
        fseries: FrequencySeries,
        *,
        norm: bool | str = True,
        epoch: float | LIGOTimeGPS | None = None,
    ) -> TimeSeries:
        """Calculate the energy `TimeSeries` for the given fseries.

        Parameters
        ----------
        fseries : `~gwpy.frequencyseries.FrequencySeries`
            The complex FFT of a time-series data set.

        norm : `bool`, `str`
            Normalize the energy of the output by the median (if `True` or
            ``'median'``) or the ``'mean'``, if `False` the output
            is the energy (power) of the Q-tranform.

        epoch : `~gwpy.time.LIGOTimeGPS`, `float`
            The epoch of these data, only used for metadata in the output
            `TimeSeries`, and not requires if the input `fseries` has the
            epoch populated.

        Returns
        -------
        energy : `~gwpy.timeseries.TimeSeries`
            A `TimeSeries` of the energy from the Q-transform of
            this tile against the data.
        """
        from ..timeseries import TimeSeries

        windowed = fseries[self.get_data_indices()] * self.get_window()
        # pad data, move negative frequencies to the end, and IFFT
        padded = numpy.pad(
            windowed,
            self.padding,
            mode="constant",
        )
        wenergy = npfft.ifftshift(padded)
        # return a `TimeSeries`
        if epoch is None:
            epoch = fseries.epoch
        tdenergy = npfft.ifft(wenergy)
        cenergy = TimeSeries(
            tdenergy,
            x0=epoch,
            dx=self.duration / tdenergy.size,
            copy=False,
        )
        energy = type(cenergy)(
            cenergy.value.real ** 2. + cenergy.value.imag ** 2.,
            x0=cenergy.x0,
            dx=cenergy.dx,
            copy=False,
        )

        if norm:
            norm = norm.lower() if isinstance(norm, str) else norm
            if norm in (True, "median"):
                narray = energy / energy.median()
            elif norm in ("mean",):
                narray = energy / energy.mean()
            else:
                msg = f"invalid normalisation '{norm}'"
                raise ValueError(msg)
            return narray.astype("float32", casting="same_kind", copy=False)
        return energy


class QGram:
    """Store tile energies over an irregularly gridded plane.

    Parameters
    ----------
    plane : `QPlane`
        The time-frequency plane over which to populate.

    energies : `list` of `TimeSeries`
        A list of signal energies for each row of tiles.

    search : `~gwpy.segments.Segment`
        Search window of interest to determine the loudest tile.
    """

    def __init__(
        self,
        plane: QPlane,
        energies: list[TimeSeries],
        search: tuple[float, float] | None,
    ) -> None:
        """Initialize a `QGram` object."""
        self.plane = plane
        self.energies = energies
        self.peak = self._find_peak(search)

    def _find_peak(self, search: tuple[float, float] | None) -> dict[str, float]:
        peak: dict[str, float] = {
            "energy": 0.,
            "snr": 0.,
            "time": 0.,
            "frequency": 0.,
        }
        for freq, energy in zip(self.plane.frequencies, self.energies, strict=True):
            if search is not None:
                energy = energy.crop(*search)  # noqa: PLW2901
            maxidx = energy.value.argmax()
            maxe = energy.value[maxidx]
            if maxe > peak["energy"]:
                peak.update({
                    "energy": maxe,
                    "snr": (2 * maxe) ** (1 / 2.),
                    "time": energy.t0.value + energy.dt.value * maxidx,
                    "frequency": freq,
                })
        return peak

    def interpolate(
        self,
        tres: float | None | str = "<default>",
        fres: float | None | str = "<default>",
        *,
        logf: bool = False,
        outseg: Segment | None = None,
    ) -> Spectrogram:
        """Interpolate this `QGram` over a regularly-gridded spectrogram.

        Parameters
        ----------
        tres : `float`
            Desired time resolution (seconds) of output `Spectrogram`,
            default is `abs(outseg) / 1000.`

        fres : `float`, `int`, `None`
            Desired frequency resolution (Hertz) of output `Spectrogram`,
            or, if ``logf=True``, the number of frequency samples;
            give `None` to skip this step and return the original resolution,
            default is 0.5 Hz or 500 frequency samples.

        logf : `bool`
            Boolean switch to enable (`True`) or disable (`False`) use of
            log-sampled frequencies in the output `Spectrogram`.

        outseg : `~gwpy.segments.Segment`
            GPS `[start, stop)` segment for output `Spectrogram`,
            default is the full duration of the input.

        Returns
        -------
        out : `~gwpy.spectrogram.Spectrogram`
            Output `Spectrogram` of normalised Q energy.

        See Also
        --------
        scipy.interpolate
            This method uses `~scipy.interpolate.InterpolatedUnivariateSpline`
            to cast all frequency rows to a common time-axis, and then
            `~scipy.interpolate.interp2d` to apply the desired frequency
            resolution across the band.

        Notes
        -----
        This method will return a `Spectrogram` of dtype ``float32`` if
        ``norm`` is given, and ``float64`` otherwise.

        To optimize plot rendering with `~matplotlib.axes.Axes.pcolormesh`,
        the output `~gwpy.spectrogram.Spectrogram` can be given a log-sampled
        frequency axis by passing `logf=True` at runtime. The `fres` argument
        is then the number of points on the frequency axis. Note, this is
        incompatible with `~matplotlib.axes.Axes.imshow`.

        It is also highly recommended to use the `outseg` keyword argument
        when only a small window around a given GPS time is of interest.
        """
        from scipy.interpolate import (
            InterpolatedUnivariateSpline,
            RectBivariateSpline,
        )

        from ..spectrogram import Spectrogram

        if outseg is None:
            outseg = self.energies[0].span
        frequencies = self.plane.frequencies
        dtype = self.energies[0].dtype

        # build regular Spectrogram from peak-Q data by interpolating each
        # (Q, frequency) `TimeSeries` to have the same time resolution
        if tres == "<default>":
            tres = abs(Segment(outseg)) / 1000.
        xout: numpy.ndarray[tuple[int], numpy.dtype[numpy.float64]] = numpy.arange(
            outseg[0],
            outseg[1],
            step=tres,
        )
        nx = xout.size
        ny = frequencies.size
        out = Spectrogram(
            numpy.empty((nx, ny), dtype=dtype),
            t0=outseg[0],
            dt=tres,
            frequencies=frequencies,
        )
        # record Q in output
        out.q = self.plane.q
        # interpolate rows
        for i, row in enumerate(self.energies):
            xrow = numpy.arange(
                row.x0.value,
                (row.x0 + row.duration).value,
                row.dx.value,
            )
            interp = InterpolatedUnivariateSpline(xrow, row.value)
            out[:, i] = interp(xout).astype(
                dtype,
                casting="same_kind",
                copy=False,
            )
        if fres is None:
            return out

        # interpolate the spectrogram to increase its frequency resolution
        # --- this is done because Duncan doesn't like interpolated images
        #     since they don't support log scaling
        interp = RectBivariateSpline(xout, frequencies, out.value)

        if not logf:
            if fres == "<default>":
                fres = .5
            outfreq = numpy.arange(
                self.plane.frange[0],
                self.plane.frange[1],
                fres,
                dtype=dtype,
            )
        else:
            if fres == "<default>":
                fres = 500
            outfreq = numpy.geomspace(
                self.plane.frange[0],
                self.plane.frange[1],
                num=int(fres),
            )
        new = type(out)(
            interp(xout, outfreq).astype(
                dtype,
                casting="same_kind",
                copy=False,
            ),
            t0=outseg[0],
            dt=tres,
            frequencies=outfreq,
        )
        new.q = self.plane.q
        return new

    def table(self, snrthresh: float = 5.5) -> EventTable:
        """Represent this `QPlane` as an `EventTable`.

        Parameters
        ----------
        snrthresh : `float`
            Lower inclusive threshold on individual tile SNR to keep in the
            table, default: 5.5.

        Returns
        -------
        out : `~gwpy.table.EventTable`
            A table of time-frequency tiles on this `QPlane`.

        Notes
        -----
        Only tiles with signal energy greater than or equal to
        `snrthresh ** 2 / 2` will be stored in the output `EventTable`.
        """
        from ..table import EventTable

        # get plane properties
        freqs = self.plane.frequencies
        bws = 2 * (freqs - self.plane.farray)

        # collect table data as a recarray
        names = (
            "time",
            "frequency",
            "duration",
            "bandwidth",
            "energy",
        )
        rec = numpy.recarray((0,), names=names, formats=["f8"] * len(names))
        for f, bw, row in zip(freqs, bws, self.energies, strict=True):
            ind, = (row.value >= snrthresh ** 2 / 2.).nonzero()
            new = ind.size
            if new > 0:
                rec.resize((rec.size + new,), refcheck=False)
                rec["time"][-new:] = row.times.value[ind]
                rec["frequency"][-new:] = f
                rec["duration"][-new:] = row.dt.to("s").value
                rec["bandwidth"][-new:] = bw
                rec["energy"][-new:] = row.value[ind]

        # save to a table
        out = EventTable(rec, copy=False)
        out.meta["q"] = self.plane.q
        return out


# -- utilities -----------------------

def q_scan(
    data: TimeSeries,
    mismatch: float = DEFAULT_MISMATCH,
    qrange: tuple[float, float] = DEFAULT_QRANGE,
    frange: tuple[float, float] = DEFAULT_FRANGE,
    duration: float | None = None,
    sampling: float | None = None,
    **kwargs,
) -> tuple[QGram, float]:
    """Transform data by scanning over a `QTiling`.

    This utility is provided mainly to allow direct manipulation of the
    `QTiling.transform` output. Most users probably just want to use
    :meth:`~gwpy.timeseries.TimeSeries.q_transform`, which wraps around this.

    Parameters
    ----------
    data : `~gwpy.timeseries.TimeSeries` or `ndarray`
        The time- or frequency-domain input data.

    mismatch : `float`
        Maximum allowed fractional mismatch between neighbouring tiles.

    qrange : `tuple` of `float`
        ``(low, high)`` range of Qs to scan.

    frange : `tuple` of `float`
        ``(low, high)`` range of frequencies to scan.

    duration : `float`
        Duration (seconds) of input, required if `data` is not a `TimeSeries`.

    sampling : `float`
        Sample rate (Hertz) of input, required if `data` is not a `TimeSeries`.

    kwargs
        Other keyword arguments to be passed to :meth:`QTiling.transform`,
        including ``'epoch'`` and ``'search'``

    Returns
    -------
    qgram : `QGram`
        The raw output of :meth:`QTiling.transform`.

    far : `float`
        Expected false alarm rate (Hertz) of white Gaussian noise with the
        same peak energy and total duration as `qgram`.
    """
    from gwpy.timeseries import TimeSeries

    # prepare input
    if isinstance(data, TimeSeries):
        duration = abs(data.span)
        sampling = data.sample_rate.to("Hz").value
        kwargs.setdefault("epoch", data.x0.value)
        data = data.fft().value

    # return a raw Q-transform and its significance
    return QTiling(
        duration,
        sampling,
        mismatch=mismatch,
        qrange=qrange,
        frange=frange,
    ).transform(data, **kwargs)



# Copyright (c) 2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read/write segments and flags from DQSEGDB-format JSON."""

import json
from typing import IO

from ...io.registry import (
    default_registry,
    identify_factory,
)
from ...io.utils import (
    with_open,
)
from .. import DataQualityFlag

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


# -- read ----------------------------

@with_open
def read_json_flag(fobj: IO) -> DataQualityFlag:
    """Read a `DataQualityFlag` from a segments-web.ligo.org JSON file."""
    data = json.load(fobj)

    # format flag
    name = "{ifo}:{name}:{version}".format(**data)
    out = DataQualityFlag(name, active=data["active"],
                          known=data["known"])

    # parse 'metadata'
    try:
        out.description = data["metadata"].get("flag_description", None)
    except KeyError:  # no metadata available, but that's ok
        pass
    else:
        out.isgood = not data["metadata"].get(
            "active_indicates_ifo_badness", False)

    return out


# -- write ---------------------------

@with_open(mode="w", pos=1)
def write_json_flag(
    flag: DataQualityFlag,
    fobj: IO,
    **kwargs,
):
    """Write a `DataQualityFlag` to a JSON file.

    Parameters
    ----------
    flag : `DataQualityFlag`
        Data to write.

    fobj : `file`
        Target file (or filename) to write.

    **kwargs
        other keyword arguments to pass to :func:`json.dump`

    See Also
    --------
    json.dump
        for details on acceptable keyword arguments
    """
    # build json packet
    data = {}
    data["ifo"] = flag.ifo
    data["name"] = flag.tag
    data["version"] = flag.version
    data["active"] = flag.active
    data["known"] = flag.known
    data["metadata"] = {}
    data["metadata"]["active_indicates_ifo_badness"] = not flag.isgood
    data["metadata"]["flag_description"] = flag.description

    # write
    json.dump(data, fobj, **kwargs)


# -- register ------------------------

default_registry.register_reader("json", DataQualityFlag, read_json_flag)
default_registry.register_writer("json", DataQualityFlag, write_json_flag)
default_registry.register_identifier(
    "json",
    DataQualityFlag,
    identify_factory("json"),
)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""The `Series` is a one-dimensional array with metadata."""

from __future__ import annotations

import contextlib
from math import floor
from typing import (
    TYPE_CHECKING,
    cast,
    overload,
)
from warnings import warn

import numpy
from astropy.units import (
    Quantity,
    Unit,
    second,
)

from ..io.registry import UnifiedReadWriteMethod
from . import sliceutils
from .array import Array
from .connect import (
    SeriesRead,
    SeriesWrite,
)
from .index import Index

if TYPE_CHECKING:
    from typing import (
        ClassVar,
        Literal,
        Self,
    )

    from astropy.units import UnitBase
    from astropy.units.typing import QuantityLike

    from ..plot import Plot
    from ..segments import Segment
    from ..typing import UnitLike
    from .sliceutils import SliceLike

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


class Series(Array):
    """A one-dimensional data series.

    A `Series` is defined as an array of data indexed upon an axis, meaning
    each sample maps to a position upon the axis. By convention the X axis
    is used to define the index, with the `~Series.x0`, `~Series.dx`, and
    `~Series.xindex` attributes allowing the positions of the data to be
    well defined.

    Parameters
    ----------
    value : array-like
        Input data array.

    unit : `~astropy.units.Unit`, optional
        Physical unit of these data.

    x0 : `float`, `~astropy.units.Quantity`, optional, default: `0`
        The starting value for the x-axis of this array.

    dx : `float`, `~astropy.units.Quantity, optional, default: `1`
        The step size for the x-axis of this array.

    xindex : `array-like`
        The complete array of x-axis values for this array.
        This argument takes precedence over `x0` and `dx` so should be
        given in place of these if relevant, not alongside.

    xunit : `~astropy.units.Unit`, optional
        The unit of the x-axis index. If not given explicitly, it will be
        taken from any of `dx`, `x0`, or `xindex`, or set to a boring default.

    epoch : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS epoch associated with these data,
        any input parsable by `~gwpy.time.to_gps` is fine.

    name : `str`, optional
        Descriptive title for this array.

    channel : `~gwpy.detector.Channel`, `str`, optional
        Source data stream for these data.

    dtype : `~numpy.dtype`, optional
        Input data type.

    copy : `bool`, optional, default: `False`
        Choose to copy the input data to new memory.

    subok : `bool`, optional, default: `True`
        Allow passing of sub-classes by the array generator.

    Returns
    -------
    series : `Series`
        a new `Series`

    Examples
    --------
    To define a `Series` of displacements at a given input laser power,
    for example:

    >>> data = Series([1, 2, 3, 2, 4, 3], unit='nm', x0=0, dx=2, xunit='W')
    >>> print(data)
    Series([ 1., 2., 3., 2., 4., 3.]
           unit: Unit("nm"),
           name: None,
           epoch: None,
           channel: None,
           x0: 0.0 W,
           dx: 2.0 W,
           xindex: [  0.   2.   4.   6.   8.  10.] W)
    """

    _metadata_slots: ClassVar[tuple[str, ...]] = (
        *Array._metadata_slots,
        "x0",
        "dx",
        "xindex",
    )
    _default_xunit: ClassVar[UnitBase] = Unit("")
    _ndim: ClassVar[int] = 1

    def __new__(
        cls,
        value: QuantityLike,
        unit: UnitBase | str | None = None,
        x0: QuantityLike | None = None,
        dx: QuantityLike | None = None,
        xindex: QuantityLike | None = None,
        xunit: UnitBase | str | None = None,
        **kwargs,
    ) -> Self:
        """Create a new `Series."""
        # check input data dimensions are OK
        shape = numpy.shape(value)
        if len(shape) != cls._ndim:
            msg = (
                f"cannot generate {cls.__name__} with "
                f"{len(shape)}-dimensional data"
            )
            raise ValueError(msg)

        # create new object
        new = super().__new__(
            cls,
            value,
            unit=unit,
            **kwargs,
        )

        # set x-axis metadata from xindex
        if xindex is not None:

            if len(xindex) != len(value):
                msg = "xindex must have the same length as data"
                raise ValueError(msg)

            # warn about duplicate settings
            if dx is not None:
                warn(
                    f"xindex was given to {cls.__name__}(), "
                    "dx will be ignored",
                    stacklevel=2,
                )
            if x0 is not None:
                warn(
                    f"xindex was given to {cls.__name__}(), "
                    "x0 will be ignored",
                    stacklevel=2,
                )
            # get unit
            if xunit is None and isinstance(xindex, Quantity):
                xunit = xindex.unit
            elif xunit is None:
                xunit = cls._default_xunit
            new.xindex = Quantity(xindex, unit=xunit)
        # or from x0 and dx
        else:
            if xunit is None and isinstance(dx, Quantity):
                xunit = dx.unit
            elif xunit is None and isinstance(x0, Quantity):
                xunit = x0.unit
            elif xunit is None:
                xunit = cls._default_xunit
            if dx is not None:
                new.dx = Quantity(dx, xunit)
            if x0 is not None:
                new.x0 = Quantity(x0, xunit)

        return new

    # -- series creation -------------

    def __array_finalize__(self, obj: Self | None) -> None:
        """Finalize the array after creation."""
        super().__array_finalize__(obj)

        # Array.__array_finalize__ might set _xindex to None, so delete it
        if getattr(self, "_xindex", None) is None:
            del self.xindex

    # -- series properties -----------

    def _update_index(
        self,
        axis: str,
        attr: str,
        value: QuantityLike,
    ) -> None:
        """Update the current axis index based on a given attr or value.

        This is an internal method designed to set the origin or step for
        an index, whilst updating existing Index arrays as appropriate.

        Parameters
        ----------
        axis : `str`
            The name of the axis to update, e.g. ``"x"``.

        attr : `str`
            The name of the attribute to set.

        value : `Quantity`, `float`
            The value to set for ``attr``.

        Examples
        --------
        >>> self._update_index("x", "x0", 0)
        >>> self._update_index("x", "dx", 1)

        Notes
        -----
        To actually set an index array, use `_set_index`.
        """
        index = f"{axis[0]}index"

        # delete current value if given None
        if value is None:
            delattr(self, attr)
            delattr(self, index)
            return

        _attr = f"_{attr}"
        unit = f"{axis[0]}unit"

        # convert float to Quantity
        if not isinstance(value, Quantity):
            try:
                value = Quantity(value, getattr(self, unit))
            except TypeError:
                value = Quantity(float(value), getattr(self, unit))

        # if value is changing, delete current index
        try:
            curr = getattr(self, _attr)
        except AttributeError:
            delattr(self, index)
        else:
            if (
                value is None
                or getattr(self, attr) is None
                or not value.unit.is_equivalent(curr.unit)
                or value != curr
            ):
                delattr(self, index)

        # set new value
        setattr(self, _attr, value)

    def _set_index(
        self,
        axis: Literal["x", "y", "z"],
        index: QuantityLike,
    ) -> None:
        """Set a new index array for this series.

        Parameters
        ----------
        axis : `str`
            The name of the axis to update, e.g. ``"x"``.

        index : `Quantity`, `numpy.ndarray`
            The index array to apply.

        Examples
        --------
        >>> self._set_index("x", numpy.arange(100))
        """
        attr = f"{axis}index"

        # if given None, delete the current index
        if index is None:
            delattr(self, attr)
            return

        origin = f"{axis}0"
        delta = f"d{axis}"

        # format input as an Index array
        if not isinstance(index, Index):
            try:
                unit = index.unit
            except AttributeError:
                unit = getattr(self, f"_default_{axis}unit")
            index = Index(index, unit=unit, copy=False)

        # reset other axis attributes
        if index.size:
            setattr(self, origin, index[0])
        if index.size > 1:
            if index.regular:  # delta will reset from index
                setattr(self, delta, index[1] - index[0])
            else:
                delattr(self, delta)

        # update index array
        setattr(self, f"_{attr}", index)

    def _index_span(self, axis: str) -> Segment:
        """Return the span of the given axis.

        Parameters
        ----------
        axis : `str`
            One of 'x', 'y', or 'z'.

        Returns
        -------
        span : `~gwpy.segments.Segment`
            The ``[start, stop)`` span of this series.
        """
        from ..segments import Segment

        axisidx = ("x", "y", "z").index(axis)
        unit = getattr(self, f"{axis}unit")

        try:
            delta = getattr(self, f"d{axis}").to(unit).value
        except AttributeError:  # irregular xindex
            index = getattr(self, f"{axis}index")
            try:
                delta = index.value[-1] - index.value[-2]
            except IndexError as exc:
                msg = (
                    "cannot determine x-axis stride (dx)"
                    "from a single data point"
                )
                raise ValueError(msg) from exc
            return Segment(index.value[0], index.value[-1] + delta)

        origin = getattr(self, f"{axis}0").to(unit).value
        return Segment(origin, origin + self.shape[axisidx] * delta)

    # x0
    @property
    def x0(self) -> Quantity:
        """X-axis coordinate of the first data point."""
        self._x0: Quantity
        try:
            return self._x0
        except AttributeError:
            try:
                self._x0 = self._xindex[0]
            except (AttributeError, IndexError):
                self._x0 = Quantity(0, self.xunit)
            return self._x0

    @x0.setter
    def x0(self, value: QuantityLike) -> None:
        self._update_index("x", "x0", value)

    @x0.deleter
    def x0(self) -> None:
        with contextlib.suppress(AttributeError):
            del self._x0

    # dx
    @property
    def dx(self) -> Quantity:
        """X-axis sample separation."""
        self._dx: Quantity
        try:
            return self._dx
        except AttributeError:
            try:
                self._xindex  # noqa: B018
            except AttributeError:
                self._dx = Quantity(1, self.xunit)
            else:
                if not self.xindex.regular:
                    msg = (
                        "this series has an irregular x-axis index, "
                        "so 'dx' is not well defined"
                    )
                    raise AttributeError(msg)
                self._dx = self.xindex[1] - self.xindex[0]
            return self._dx

    @dx.setter
    def dx(self, value: QuantityLike) -> None:
        self._update_index("x", "dx", value)

    @dx.deleter
    def dx(self) -> None:
        with contextlib.suppress(AttributeError):
            del self._dx

    # xindex
    @property
    def xindex(self) -> Index:
        """Positions of the data on the x-axis."""
        self._xindex: Index
        try:
            return self._xindex
        except AttributeError:
            self._xindex = Index.define(self.x0, self.dx, self.shape[0])
            return self._xindex

    @xindex.setter
    def xindex(self, index: QuantityLike) -> None:
        self._set_index("x", index)

    @xindex.deleter
    def xindex(self) -> None:
        with contextlib.suppress(AttributeError):
            del self._xindex

    # xunit
    @property
    def xunit(self) -> UnitBase:
        """Unit of x-axis index."""
        try:
            return self._dx.unit
        except AttributeError:
            try:
                return self._x0.unit
            except AttributeError:
                return self._default_xunit

    @xunit.setter
    def xunit(self, unit: UnitLike) -> None:
        unit = Unit(unit)
        try:  # set the index, if present
            self.xindex = self._xindex.to(unit)
        except AttributeError:  # or just set the start and step
            self.dx = self.dx.to(unit)
            self.x0 = self.x0.to(unit)

    @property
    def xspan(self) -> Segment:
        """X-axis [low, high) segment encompassed by these data."""
        return self._index_span("x")

    # -- series i/o ------------------

    read = UnifiedReadWriteMethod(SeriesRead)
    write = UnifiedReadWriteMethod(SeriesWrite)

    # -- series plotting -------------

    def plot(
        self,
        method: str = "plot",
        **kwargs,
    ) -> Plot:
        """Plot the data for this series.

        Parameters
        ----------
        method : `str`, optional
            The method on the `~gwpy.plot.Axes` to call to render
            this object. Default is ``"plot"`` (`~gwpy.plot.Axes.plot`).

        kwargs
            Other keyword arguments are passed to the relevant
            Axes plotting method.

        Returns
        -------
        figure : `~matplotlib.figure.Figure`
            The newly created figure, with populated Axes.

        See Also
        --------
        matplotlib.pyplot.figure
            For documentation of keyword arguments used to create the figure.

        matplotlib.figure.Figure.add_subplot
            For documentation of keyword arguments used to create the axes.

        matplotlib.axes.Axes.plot
            For documentation of keyword arguments used in rendering the data.
        """
        from ..plot import Plot
        from ..plot.text import default_unit_label

        # correct for log scales and zeros
        if kwargs.get("xscale") == "log" and self.x0.value == 0:
            kwargs.setdefault("xlim", (self.dx.value, self.xspan[1]))

        # make plot
        plot = Plot(self, method=method, **kwargs)

        # set default y-axis label (xlabel is set by Plot())
        default_unit_label(plot.gca().yaxis, self.unit)

        return plot

    def step(self, **kwargs) -> Plot:
        """Create a step plot of this series.

        kwargs
            All keyword arguments are passed to the :meth:`plot` method.
            of this series.

        See Also
        --------
        plot
            For details of the plotting.
        """
        where = kwargs.pop("where", "post")
        kwargs.setdefault(
            "drawstyle",
            f"steps-{where}",
        )
        data = self.append(self.value[-1:], inplace=False)
        return data.plot(**kwargs)

    # -- series methods --------------

    def shift(self, delta: QuantityLike) -> None:
        """Shift this `Series` forward on the X-axis by ``delta``.

        This modifies the series in-place.

        Parameters
        ----------
        delta : `float`, `~astropy.units.Quantity`, `str`
            The amount by which to shift (in x-axis units if `float`), give
            a negative value to shift backwards in time

        Examples
        --------
        >>> from gwpy.types import Series
        >>> a = Series([1, 2, 3, 4, 5], x0=0, dx=1, xunit='m')
        >>> print(a.x0)
        0.0 m
        >>> a.shift(5)
        >>> print(a.x0)
        5.0 m
        >>> a.shift('-1 km')
        -995.0 m
        """
        self.x0 = self.x0 + Quantity(delta, self.xunit)

    def value_at(self, x: QuantityLike) -> Quantity:
        """Return the value of this `Series` at the given `xindex` value.

        Parameters
        ----------
        x : `float`, `~astropy.units.Quantity`
            The `xindex` value at which to search.

        Returns
        -------
        y : `~astropy.units.Quantity`
            The value of this Series at the given `xindex` value.

        Raises
        ------
        IndexError
            If ``x`` doesn't match an X-index value.
        """
        x = Quantity(x, self.xindex.unit).value
        try:
            idx = (self.xindex.value == x).nonzero()[0][0]
        except IndexError as e:
            e.args = (f"Value {x!r} not found in array index",)
            raise
        return self[idx]

    def copy(self, order: Literal["C", "F", "A", "K"] = "C") -> Self:
        """Return a copy of this `Series`.

        Parameters
        ----------
        order : {'C', 'F', 'A', 'K'}, optional
            The desired memory layout order for the copy.

        See Also
        --------
        numpy.ndarray.copy
            For details of the copy operation.
        """
        new = super().copy(order=order)
        with contextlib.suppress(AttributeError):
            new._xindex = self._xindex.copy()
        return new

    copy.__doc__ = Array.copy.__doc__

    def zip(self) -> numpy.ndarray:
        """Zip the `xindex` and `value` arrays of this `Series`.

        Returns
        -------
        stacked : 2-d `numpy.ndarray`
            The array formed by stacking the the `xindex` and `value` of this
            series

        Examples
        --------
        >>> a = Series([0, 2, 4, 6, 8], xindex=[-5, -4, -3, -2, -1])
        >>> a.zip()
        array([[-5.,  0.],
               [-4.,  2.],
               [-3.,  4.],
               [-2.,  6.],
               [-1.,  8.]])

        """
        return numpy.column_stack((self.xindex.value, self.value))

    def diff(self, n: int = 1, axis: int = -1) -> Self:
        """Calculate the n-th order discrete difference along given axis.

        The first order difference is given by ``out[n] = a[n+1] - a[n]`` along
        the given axis, higher order differences are calculated by using `diff`
        recursively.

        Parameters
        ----------
        n : int, optional
            The number of times values are differenced.

        axis : int, optional
            The axis along which the difference is taken, default is the
            last axis.

        Returns
        -------
        diff : `Series`
            The `n` order differences. The shape of the output is the same
            as the input, except along `axis` where the dimension is
            smaller by `n`.

        See Also
        --------
        numpy.diff
            For documentation on the underlying method.
        """
        out = super().diff(n=n, axis=axis)
        try:
            out.x0 = self.x0 + self.dx * n
        except AttributeError:  # irregular xindex
            out.x0 = self.xindex[n]
        return out

    @overload
    def __getitem__(self, item: int) -> Quantity: ...
    @overload
    def __getitem__(self, item: SliceLike | tuple[SliceLike, ...]) -> Self: ...

    def __getitem__(
        self,
        item: SliceLike | tuple[SliceLike, ...],
    ) -> Self | Quantity:
        """Get an item, or a slice, from this `Series`."""
        new = super().__getitem__(item)

        # slice axis 0 metadata
        (slice_,) = sliceutils.format_nd_slice(item, 1)
        if not sliceutils.null_slice(slice_):
            sliceutils.slice_axis_attributes(self, "x", new, "x", slice_)

        return new

    def is_contiguous(
        self,
        other: Series | numpy.ndarray | list,
        tol: float = 2 ** -18,
    ) -> int:
        """Check whether other is contiguous with self.

        Parameters
        ----------
        other : `Series`, `numpy.ndarray`
            Another series of the same type to test for contiguity.

        tol : `float`, optional
            The numerical tolerance of the test.

        Returns
        -------
        1
            If ``other`` is contiguous with this series,
            i.e. would attach seamlessly onto the end.

        -1
            If ``other`` is anti-contiguous with this seires,
            i.e. would attach seamlessly onto the start.

        0
            If ``other`` is completely dis-contiguous with this series.

        Notes
        -----
        If ``other`` is an array that doesn't have index information (e.g.
        a `numpy.ndarray`), this method always returns ``1``.

        If ``self`` ***or*** ``other``` have an irregular `Index` array
        (e.g. aren't linearly sampled), this method will always return ``1``
        if ``other`` starts after ``self`` finishes, or ``-1``` if the inverse.
        If the two arrays overlap, that is bad and will raise an error.
        """
        self.is_compatible(other)
        if isinstance(other, type(self)):
            if abs(float(self.xspan[1] - other.xspan[0])) < tol:
                return 1
            if abs(float(other.xspan[1] - self.xspan[0])) < tol:
                return -1
            return 0
        if type(other) in [list, tuple, numpy.ndarray]:
            return 1
        return 0

    def is_compatible(self, other: list | numpy.ndarray) -> bool:
        """Check whether this series and other have compatible metadata.

        This method tests that the `sample size <Series.dx>`, and the
        `~Series.unit` match.
        """
        self.check_compatible(other)
        # we survived!
        return True

    def check_compatible(
        self,
        other: list | numpy.ndarray,
        casting: Literal["no", "equiv", "safe", "same_kind", "unsafe"] | None = "safe",
        *,
        irregular_equal: bool = True,
    ) -> None:
        """Check whether this Series and ``other`` are compatible.

        Parameters
        ----------
        other : `numpy.ndarray`, `Series`
            The array to compare to.

        casting : `str`, optional
            The type of casting to support when comparing dtypes.

        irregular_equal : `bool`, optional
            Require irregular indices to be equal (default).
            If ``irregular_equal=False`` and either (or both) of the series
            are irregular, this method just returns without doing anything.

        Raises
        ------
        ValueError
            If any metadata elements aren't compatible.

        TypeError
            If the dtype can't be safely cast between the arrays.
        """
        if isinstance(other, type(self)):
            self._check_compatible_gwpy(
                other,
                irregular_equal=irregular_equal,
            )
        if isinstance(other, Quantity):
            self._check_compatible_quantity(other)
        self._check_compatible_numpy(other, casting=casting)

    def _compatibility_error(
        self,
        other: Series,
        attr: str,
        desc: str,
    ) -> ValueError:
        """Construct a `ValueError` to be raised somewhere else."""
        msg = (
            f"{type(self).__name__} {desc} do not match: "
            f"{getattr(self, attr, None)} vs "
            f"{getattr(other, attr, None)}"
        )
        return ValueError(msg)

    def _check_compatible_index(
        self,
        other: Series,
        axis: str = "x",
        *,
        irregular_equal: bool = True,
    ) -> None:
        """Compare index attributes/arrays between self and other.

        Parameters
        ----------
        other : `Series`
            The series to compare to this one.

        axis : `str`, optional
            The Axis index to compare.
            Default is ``x``.

        irregular_equal : `bool`, optional
            Require irregular indices to be equal (default).
            If ``irregular_equal=False`` and either (or both) of the series
            are irregular, this method just returns without doing anything.

        Raises
        ------
        ValueError
            If ``dx`` doesn't match, or ``xindex`` values are not present/are
            identical (as appropriate).
        """
        _delta = f"d{axis}"
        try:  # check step size, if possible
            deltaa = getattr(self, _delta)
            deltab = getattr(other, _delta)
        except AttributeError:  # irregular index
            # at least one of the series is irregular
            _index = f"_{axis}index"
            idxa = getattr(self, _index, None)
            idxb = getattr(other, _index, None)

            # compare the indices
            if irregular_equal and (
                idxa is None  # no index on 'self'
                or idxb is None  # no index on 'other'
                or not numpy.array_equal(idxa, idxb)  # indexes don't match
            ):
                raise self._compatibility_error(
                    other,
                    _index,
                    f"{axis}-axis indexes",
                ) from None
        else:
            # both series are regular, check that the step sizes match
            if deltaa != deltab:
                raise self._compatibility_error(
                    other,
                    _delta,
                    f"{axis}-axis sample sizes",
                )

    def _check_compatible_gwpy(
        self,
        other: Quantity,
        *,
        irregular_equal: bool = True,
    ) -> None:
        """Check whether this series and another series are compatible.

        This method checks that the Index arrays are compatible.
        """
        self._check_compatible_index(
            other,
            axis="x",
            irregular_equal=irregular_equal,
        )

    def _check_compatible_quantity(
        self,
        other: Quantity,
    ) -> None:
        """Check with this Series and another `Quantity` are compatible.

        This method checks that the units match.

        Raises
        ------
        ValueError
            If this series and ``other`` are incompatible.

        See Also
        --------
        astropy.units.Unit.is_equivalent
            For details of the unit compatibility check.
        """
        # if neither quantity has a unit, that's fine
        if self.unit is None and other.unit is None:
            return
        # check that the units are equivalent
        if (
            self.unit is None
            or other.unit is None
            or not self.unit.is_equivalent(other.unit)
        ):
            raise self._compatibility_error(other, "unit", "units")

    def _check_compatible_numpy(
        self,
        other: list | numpy.ndarray,
        casting: Literal["no", "equiv", "safe", "same_kind", "unsafe"] | None = "safe",
    ) -> None:
        """Check whether this series and a numpy.ndarray are compatible.

        This method checks that the dimensions are the same, and that
        the dtypes match.

        Raises
        ------
        ValueError
            If the dimensions don't match.

        TypeError
            If the data types can't be cast safely.

        See Also
        --------
        numpy.can_cast
            For details of the dtype check.
        """
        arr = numpy.asarray(other)
        if arr.ndim != self.ndim:
            msg = "dimensionality does not match"
            raise ValueError(msg)
        if not numpy.can_cast(arr.dtype, self.dtype, casting=casting):
            msg = f"array data types do not match: {self.dtype} vs {arr.dtype}"
            raise TypeError(msg)

    def append(
        self,
        other: numpy.ndarray | Series,
        *,
        inplace: bool = True,
        gap: Literal["raise", "ignore", "pad"] | None = None,
        pad: float | None = None,
        resize: bool = True,
    ) -> Self:
        """Connect another series onto this one.

        Parameters
        ----------
        other : `numpy.ndarray`, `Series`
            Another `Series`, or a simple data array to connect to this one.

        inplace : `bool`, optional
            If `True` (default) perform the operation in-place, modifying current
            series.
            If `False` copy the data to new memory before modifying.

            .. warning::

               ``inplace`` append bypasses the reference check in
               `numpy.ndarray.resize`, so be carefully to only use this
               for arrays that haven't been sharing their memory!

        gap : `str`, optional
            Action to perform if there's a gap between the other series
            and this one. One of

            - ``'raise'`` - raise a `ValueError`
            - ``'ignore'`` - remove gap and join data
            - ``'pad'`` - pad gap with zeros

            If ``pad`` is given and is not `None`, the default is
            ``gap='pad'``, otherwise ``gap='raise'``.

            If ``gap='pad'`` is given, the default for ``pad`` is ``0``.

        pad : `float`, optional
            Value with which to pad discontiguous series,
            by default gaps will result in a `ValueError`.

        resize : `bool`, optional
            If `True` (default) resize this array to accommodate new data.
            If `False` roll the current data like a buffer to the left
            and insert new data at the other end.

        Returns
        -------
        series : `Series`
            A new series containing joined data sets.
        """
        if gap is None:
            gap = "raise" if pad is None else "pad"
        if pad is None and gap == "pad":
            pad = 0.0

        # check metadata
        self.is_compatible(other)
        # make copy if needed
        if not inplace:
            self = self.copy()
        # fill gap
        if self.is_contiguous(other) != 1:
            other = cast("Self", other)
            if gap == "pad":
                pad = cast("float", pad)
                ngap = floor((other.xspan[0] - self.xspan[1]) / self.dx.value + 0.5)
                if ngap < 1:
                    msg = (
                        "Cannot append {0} that starts before this one:\n"
                        "    {0} 1 span: {1}\n    {0} 2 span: {2}".format(
                            type(self).__name__,
                            self.xspan,
                            other.xspan,
                        )
                    )
                    raise ValueError(
                        msg,
                    )
                gapshape = list(self.shape)
                gapshape[0] = int(ngap)
                padding = (numpy.ones(gapshape) * pad).astype(self.dtype)
                self.append(padding, inplace=True, resize=resize)
            elif gap == "ignore":
                pass
            elif self.xspan[0] < other.xspan[0] < self.xspan[1]:
                msg = (
                    "Cannot append overlapping {0}s:\n"
                    "    {0} 1 span: {1}\n    {0} 2 span: {2}".format(
                        type(self).__name__,
                        self.xspan,
                        other.xspan,
                    )
                )
                raise ValueError(
                    msg,
                )
            else:
                msg = (
                    "Cannot append discontiguous {0}\n"
                    "    {0} 1 span: {1}\n    {0} 2 span: {2}".format(
                        type(self).__name__,
                        self.xspan,
                        other.xspan,
                    )
                )
                raise ValueError(
                    msg,
                )

        # check empty other
        if not other.size:
            return self

        # resize first
        if resize:
            N = other.shape[0]
            s = list(self.shape)
            s[0] = self.shape[0] + other.shape[0]
            try:
                self.resize(s, refcheck=False)
            except ValueError as e:
                if "resize only works on single-segment arrays" in str(e):
                    self = self.copy()
                    self.resize(s)
                else:
                    raise
        elif other.shape[0] < self.shape[0]:
            N = other.shape[0]
            self.value[:-N] = self.value[N:]
        else:
            N = min(self.shape[0], other.shape[0])

        # if units are the same, can shortcut
        if isinstance(other, Series) and other.unit == self.unit:
            self.value[-N:] = other.value[-N:]
        # otherwise if its just a numpy array
        elif type(other) is type(self.value) or (other.dtype.name.startswith("uint")):
            self.value[-N:] = other[-N:]
        else:
            self[-N:] = other[-N:]
        try:
            self._xindex  # noqa: B018
        except AttributeError:
            if not resize:
                self.x0 = self.x0.value + other.shape[0] * self.dx.value
        else:
            if resize:
                try:
                    self.xindex.resize((s[0],), refcheck=False)
                except ValueError as exc:
                    if "cannot resize" in str(exc):
                        self.xindex = self.xindex.copy()
                        self.xindex.resize((s[0],))
                    else:
                        raise
            else:
                self.xindex[: -other.shape[0]] = self.xindex[other.shape[0] :]
            try:
                self.xindex[-other.shape[0] :] = other._xindex  # type: ignore[union-attr]  # noqa: SLF001
            except AttributeError:
                del self.xindex
                if not resize:
                    self.x0 = self.x0 + self.dx * other.shape[0]
            else:
                with contextlib.suppress(IndexError):
                    self.dx = self.xindex[1] - self.xindex[0]
                self.x0 = self.xindex[0]
        return self

    def prepend(
        self,
        other: QuantityLike,
        *,
        inplace: bool = True,
        gap: Literal["raise", "ignore", "pad"] | None = None,
        pad: float | None = None,
        resize: bool = True,
    ) -> Series:
        """Connect another series onto the start of the current one.

        Parameters
        ----------
        other : `numpy.ndarray`, `Series`
            The data to prepend to this series.

        inplace : `bool`, optional
            If `True` (default) perform the operation in-place, modifying current
            series.
            If `False` copy the data to new memory before modifying.

            .. warning::

               ``inplace`` append bypasses the reference check in
               `numpy.ndarray.resize`, so be carefully to only use this
               for arrays that haven't been sharing their memory!

        gap : `str`, optional
            Action to perform if there's a gap between the other series
            and this one. One of

            - ``'raise'`` - raise a `ValueError`
            - ``'ignore'`` - remove gap and join data
            - ``'pad'`` - pad gap with zeros

            If ``pad`` is given and is not `None`, the default is
            ``gap='pad'``, otherwise ``gap='raise'``.

            If ``gap='pad'`` is given, the default for ``pad`` is ``0``.

        pad : `float`, optional
            Value with which to pad discontiguous series,
            by default gaps will result in a `ValueError`.

        resize : `bool`, optional
            If `True` (default) resize this array to accommodate new data.
            If `False` roll the current data like a buffer to the left or right
            (depending on ``prepend``) and insert new data at the other end.

        Returns
        -------
        series : `Series`
            The modified series.
        """
        out = other.append(self, inplace=False, gap=gap, pad=pad, resize=resize)
        if inplace:
            self.resize(out.shape, refcheck=False)
            self[:] = out[:]
            self.x0 = out.x0.copy()
            del out
            return self
        return out

    def update(
        self,
        other: QuantityLike,
        *,
        inplace: bool = True,
        gap: Literal["raise", "ignore", "pad"] | None = None,
        pad: float | None = None,
    ) -> Self:
        """Update this series by appending new data like a buffer.

        Old data (at the start) are dropped to maintain a fixed size.

        This is a convenience method that just calls `~Series.append` with
        ``resize=False``.

        Parameters
        ----------
        other : `Series`, `numpy.ndarray`
            The data to add to the end of this `Series`.

        inplace : `bool`
            If `True` (default) modify the data in place.
            If `False` copy the data to new memory.

        gap : `str`, optional
            Action to perform if there's a gap between the other series
            and this one. One of

            - ``'raise'`` - raise a `ValueError`
            - ``'ignore'`` - remove gap and join data
            - ``'pad'`` - pad gap with zeros

            If ``pad`` is given and is not `None`, the default is
            ``gap='pad'``, otherwise ``gap='raise'``.

            If ``gap='pad'`` is given, the default for ``pad`` is ``0``.

        pad : `float`, optional
            Value with which to pad discontiguous series,
            by default gaps will result in a `ValueError`.

        Returns
        -------
        series : `Series`
            Either the same series (if ``inplace=True``) or a new
            series (if ``inplace=False``) with ``other`` data added
            to the end of this 'buffer'.

        See Also
        --------
        append
            For details of the data manipulation.
        """
        return self.append(
            other,
            inplace=inplace,
            gap=gap,
            pad=pad,
            resize=False,
        )

    def crop(
        self,
        start: Quantity | float | None = None,
        end: Quantity | float | None = None,
        *,
        copy: bool = False,
    ) -> Self:
        """Crop this series to the given x-axis extent.

        Parameters
        ----------
        start : `float`, optional
            Lower limit of x-axis to crop to, defaults to
            :attr:`~Series.x0`.

        end : `float`, optional
            Upper limit of x-axis to crop to, defaults to series end.

        copy : `bool`, optional
            Copy the input data to fresh memory,
            otherwise return a view (default).

        Returns
        -------
        series : `Series`
            A new series with a sub-set of the input data.

        Notes
        -----
        If either ``start`` or ``end`` are outside of the original
        `Series` span, warnings will be printed and the limits will
        be restricted to the :attr:`~Series.xspan`.
        """
        x0, x1 = self.xspan
        xtype = type(x0)
        if isinstance(start, Quantity):
            start = start.to(self.xunit).value
        if isinstance(end, Quantity):
            end = end.to(self.xunit).value

        # pin early starts to time-series start
        if start == x0:
            start = None
        elif start is not None and xtype(start) < x0:
            warn(
                f"{type(self).__name__}.crop given start smaller than current "
                "start, crop will begin when the Series actually starts.",
                stacklevel=2,
            )
            start = None

        # pin late ends to time-series end
        if end == x1:
            end = None
        if end is not None and xtype(end) > x1:
            warn(
                f"{type(self).__name__}.crop given end larger than current "
                "end, crop will begin when the Series actually ends.",
                stacklevel=2,
            )
            end = None

        # check if we have an index to use when searching
        have_xindex = getattr(self, "_xindex", None) is not None

        # find start index
        if start is None:
            idx0 = None
        elif have_xindex:
            idx0 = numpy.searchsorted(
                self.xindex.value,
                xtype(start),
                side="left",
            )
        else:
            idx0 = floor((xtype(start) - x0) / self.dx.value)

        # find end index
        if end is None:
            idx1 = None
        elif have_xindex:
            idx1 = numpy.searchsorted(
                self.xindex.value,
                xtype(end),
                side="left",
            )
        else:
            idx1 = floor((xtype(end) - x0) / self.dx.value)

        # crop
        if copy:
            return self[idx0:idx1].copy()
        return self[idx0:idx1]

    def pad(
        self,
        pad_width: int | tuple[int, int],
        **kwargs,
    ) -> Self:
        """Pad this series to a new size.

        This just wraps `numpy.pad` and handles shifting the `Index` to
        accommodate padding on the left.

        Parameters
        ----------
        pad_width : `int`, `tuple[int, int]`
            Number of samples by which to pad each end of the array;
            given a single `int` to pad both ends by the same amount,
            or a ``(before, after)`` `tuple` for assymetric padding.

        kwargs
            Other keyword arguments are passed to `numpy.pad`.

        Returns
        -------
        series : `Series`
            The padded version of the input.

        See Also
        --------
        numpy.pad
            For details on the pad function and valid keyword arguments.
        """
        # format arguments
        kwargs.setdefault("mode", "constant")
        if isinstance(pad_width, int):
            pad_width = cast("tuple[int, int]", (pad_width,))
        # form pad and view to this type
        new = cast("Self", numpy.pad(self.value, pad_width, **kwargs).view(type(self)))
        # numpy.pad has stripped all metadata, so copy it over
        new.__metadata_finalize__(self)
        new._unit = self.unit
        # finally move the starting index based on the amount of left-padding
        new.x0 = new.x0 - self.dx * pad_width[0]
        return new

    def inject(
        self,
        other: Series,
        *,
        inplace: bool = False,
    ) -> Self:
        """Add two compatible `Series` along their shared x-axis values.

        Parameters
        ----------
        other : `Series`
            A `Series` whose xindex intersects with `self.xindex`.

        inplace : `bool`, optional
            If `True` (default) perform the operation in-place,
            modifying the current series.
            If `False` copy the data to new memory before modifying.

        Returns
        -------
        out : `Series`
            The sum of `self` and `other` along their shared x-axis values.

        Raises
        ------
        ValueError
            If `self` and `other` have incompatible units or xindex intervals.

        Notes
        -----
        The offset between ``self`` and ``other`` will be rounded to the nearest
        sample if they are not exactly aligned.

        If `self.xindex` is an array of timestamps, and if `other.xspan` is
        not a subset of `self.xspan`, then `other` will be cropped before
        being adding to `self`.

        Users may wish to taper or window their `Series` before
        passing it to this method. See :meth:`TimeSeries.taper` and
        :func:`~gwpy.signal.window.planck` for more information.
        """
        # check Series compatibility
        self.is_compatible(other)

        # crop to fit
        if (self.xunit == second) and (other.xspan[0] < self.xspan[0]):
            other = other.crop(start=self.xspan[0])
        if (self.xunit == second) and (other.xspan[1] > self.xspan[1]):
            other = other.crop(end=self.xspan[1])

        # find index offset
        ox0 = other.x0.to(self.x0.unit)
        idx = ((ox0 - self.x0) / self.dx).value
        if not idx.is_integer():
            warn(
                "Series have overlapping xspan but their x-axis values are not "
                "offset by an integer number of samples, "
                "will round to the nearest sample.",
                stacklevel=2,
            )
            idx = round(idx)

        # add the Series along their shared samples
        slice_ = slice(int(idx), int(idx) + other.shape[0])
        if inplace:
            out = self
        else:
            out = self.copy()
        out.value[slice_] += other.value
        return out
